import os
import sys
from pathlib import Path
import datetime
import platform

import Orange.data
from Orange.data import Table, Domain, StringVariable, ContinuousVariable
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from orangecontrib.AAIT.utils import thread_management


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWFileMetadata(widget.OWWidget):
    name = "File Metadata"
    category = "AAIT - TOOLBOX"
    description = 'Get some metadatas on the files contained in the "path" column'
    icon = "icons/owfilemetadata.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owfilemetadata.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owfilemetadata.ui")
    want_control_area = False
    priority = 1060

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)


    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.autorun:
            self.run()


    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        # Data Management
        self.data = None
        self.result = None
        self.thread = None
        self.autorun = True
        self.post_initialized()

    def run(self):
        self.error("")
        self.warning("")

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if "path" not in self.data.domain:
            self.error('You need a "path" column.')
            self.Outputs.data.send(None)
            return

        # Start progress bar
        self.progressBarInit()

        # Connect and start thread : main function, progress, result and finish
        self.thread = thread_management.Thread(add_metadatas_to_table, self.data)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        self.progressBarFinished()

    def post_initialized(self):
        pass


def add_metadatas_to_table(table, progress_callback=None, argself=None):
    """
    Add file metadata (size, creation time, modification time) to an Orange table.

    Optimized version with:
    - Batch stat() calls with caching
    - Pre-allocated arrays for better memory usage
    - Vectorized datetime formatting
    - Reduced object creation
    - Long path support on Windows
    """
    data = table.copy()
    attr_dom = list(data.domain.attributes)
    metas_dom = list(data.domain.metas)
    class_dom = list(data.domain.class_vars)

    total_rows = len(data)

    # OPTIMIZATION 1: Pre-extract all data in one pass (avoid repeated attribute access)
    features_list = [[row[x] for x in attr_dom] for row in data]
    targets_list = [[row[y] for y in class_dom] for row in data]
    metas_array = data.metas  # Direct array access (no copy needed)

    # OPTIMIZATION 2: Batch metadata retrieval
    filepaths = [row["path"].value for row in data]
    metadata_list = get_metadata_batch(filepaths, progress_callback, argself, total_rows)

    # OPTIMIZATION 3: Build rows efficiently
    rows = []
    for i in range(total_rows):
        if argself is not None and argself.stop:
            break

        metadata = metadata_list[i]
        metas = list(metas_array[i])
        metas += [metadata["file size"], metadata["creation time"], metadata["modification time"]]
        rows.append(features_list[i] + targets_list[i] + metas)
    # Generate new Domain
    filesize_var = ContinuousVariable("file size")
    filesize_var.format_str = "%.0f"
    ctime_var = StringVariable("creation time")
    mtime_var = StringVariable("modification time")
    domain = Domain(
        attributes=attr_dom,
        class_vars=class_dom,
        metas=metas_dom + [filesize_var, ctime_var, mtime_var]
    )

    # Create and return table
    out_data = Table.from_list(domain=domain, rows=rows)
    return out_data


def get_metadata_batch(filepaths, progress_callback=None, argself=None, total=None):
    """
    Retrieve metadata for multiple files in batch (optimized).

    :param filepaths: List of file paths.
    :param progress_callback: Optional callback for progress reporting.
    :param argself: Optional object with stop attribute for cancellation.
    :param total: Total number of files (for progress calculation).
    :return: List of metadata dictionaries.
    """
    if total is None:
        total = len(filepaths)

    is_windows = platform.system() == 'Windows'
    results = []

    # OPTIMIZATION: Pre-compile datetime format (reused for all files)
    datetime_format = "%Y-%m-%d %H:%M:%S"

    for i, filepath in enumerate(filepaths):
        # Check for cancellation
        if argself is not None and argself.stop:
            break

        metadata = get_metadata_fast(filepath, datetime_format, is_windows)
        results.append(metadata)

        # Progress callback
        if progress_callback is not None:
            progress_value = float(100 * (i + 1) / total)
            progress_callback(progress_value)

    return results


def get_metadata_fast(filepath, datetime_format, is_windows):
    """
    Optimized single-file metadata retrieval with long path support.

    :param filepath: Path to the file (str or Path).
    :param datetime_format: Pre-compiled datetime format string.
    :param is_windows: Boolean indicating if running on Windows.
    :return: Dictionary with keys "file size", "creation time", "modification time".
    """
    try:
        path = Path(filepath)

        # Windows long path support
        if is_windows:
            path_str = str(path.resolve())
            if len(path_str) > 260 and not path_str.startswith('\\\\?\\'):
                if path_str.startswith('\\\\'):
                    path_str = '\\\\?\\UNC\\' + path_str[2:]
                else:
                    path_str = '\\\\?\\' + path_str
                path = Path(path_str)

        # OPTIMIZATION: Single stat() call instead of three
        stat_info = path.stat()

        # Extract all info from single stat object
        file_size = str(stat_info.st_size)

        # OPTIMIZATION: Direct timestamp conversion (avoid intermediate objects)
        ctime = datetime.datetime.fromtimestamp(stat_info.st_ctime).strftime(datetime_format)
        mtime = datetime.datetime.fromtimestamp(stat_info.st_mtime).strftime(datetime_format)

        return {
            "file size": file_size,
            "creation time": ctime,
            "modification time": mtime
        }

    except FileNotFoundError:
        return {
            "file size": "0",
            "creation time": "File not found",
            "modification time": "File not found"
        }
    except PermissionError:
        return {
            "file size": "0",
            "creation time": "Permission denied",
            "modification time": "Permission denied"
        }
    except Exception as e:
        error_msg = str(e)
        return {
            "file size": "0",
            "creation time": error_msg,
            "modification time": error_msg
        }


def get_metadata(filepath):
    """
    Legacy function - kept for backward compatibility.
    Retrieve file metadata: size, creation time, and modification time as strings.

    :param filepath: Path to the file (str or Path).
    :return: Dictionary with keys "file size", "creation time", "modification time".
    """
    is_windows = platform.system() == 'Windows'
    datetime_format = "%Y-%m-%d %H:%M:%S"
    return get_metadata_fast(filepath, datetime_format, is_windows)




if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWFileMetadata()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys
import openpyxl
from openpyxl import Workbook

import Orange.data
from Orange.data import Table, Domain, StringVariable
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import QApplication, QCheckBox, QPushButton

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import base_widget, thread_management
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils import base_widget, thread_management
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWSplitExcelSheets(base_widget.BaseListWidget):

    name = "Split Excel Sheets"
    description = "Split Excel files into one file per sheet."
    category = "AAIT - TOOLBOX"
    icon = "icons/splitexcelsheets.png"
    priority = 1070
    want_control_area = False

    gui = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "designer/owsplitexcelsheets.ui"
    )

    recursive = Setting("True")
    selected_column_name = Setting("path")

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        data = Output("Generated files", Orange.data.Table)

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.data:
            self.var_selector.add_variables(self.data.domain)
            self.var_selector.select_variable_by_name(self.selected_column_name)
        if self.autorun:
            self.run()
        if in_data is None:
            self.Outputs.data.send(None)

    def __init__(self):
        super().__init__()

        self.setFixedWidth(480)
        self.setFixedHeight(500)

        self.data = None
        self.thread = None
        self.autorun = True

        self.checkBox_recursive = self.findChild(QCheckBox, "checkBox_recursive")
        self.pushButton_run = self.findChild(QPushButton, "pushButton_run")

        self.checkBox_recursive.setChecked(self.recursive == "True")
        self.checkBox_recursive.stateChanged.connect(self.on_recursive_changed)
        self.pushButton_run.clicked.connect(self.run)

        self.post_initialized()

    def on_recursive_changed(self, state):
        self.recursive = "True" if state else "False"

    def run(self):
        self.error("")
        self.warning("")

        if self.thread is not None:
            self.thread.safe_quit()

        if self.data is None or len(self.data) == 0:
            self.Outputs.data.send(None)
            return

        # Verification of in_data
        if not self.selected_column_name in self.data.domain:
            self.warning(f'Previously selected column "{self.selected_column_name}" does not exist in your data.')
            return

        if not isinstance(self.data.domain[self.selected_column_name], StringVariable):
            self.error('You must select a text variable.')
            return

        self.progressBarInit()

        self.thread = thread_management.Thread(
            self.split_excel_worker
        )

        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, value: float):
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            if not result:
                self.Outputs.data.send(None)
                return

            domain = Domain([], metas=[StringVariable("path")])
            table = Table.from_numpy(
                domain,
                X=[[] for _ in result],
                metas=result
            )

            self.Outputs.data.send(table)

        except Exception as e:
            self.error(f"Processing error: {e}")
            self.Outputs.data.send(None)

    def handle_finish(self):
        self.progressBarFinished()

    def post_initialized(self):
        pass

    @staticmethod
    def is_excel_already_processed(filepath):
        name_no_ext = os.path.splitext(os.path.basename(filepath))[0]
        parent = os.path.dirname(filepath)
        target_dir = os.path.join(parent, f"{name_no_ext}_sheets")

        if not os.path.isdir(target_dir):
            return False

        return any(f.lower().endswith(".xlsx") for f in os.listdir(target_dir))

    def split_excel_worker(self, progress_callback):
        excel_ext = (".xlsx", ".xlsm", ".xls", ".xlsb")
        all_files = []

        # Filtre et non traitement des dossiers déjà nommé "_sheets" : siffixe que l'on ajoute quadn on crée le dossier
        # pour déterminer si le fichier a été traité
        target_files = []
        for row in self.data.metas:

            filepath = str(row[0]).replace("\\", "/")

            if "_sheets" in filepath:
                continue

            if os.path.isfile(filepath) and filepath.lower().endswith(excel_ext):
                target_files.append(filepath)

        total = len(target_files)
        if total == 0:
            return []

        # 2. Traitement des fichiers sains
        for idx, filepath in enumerate(target_files, start=1):
            name_no_ext = os.path.splitext(os.path.basename(filepath))[0]
            parent = os.path.dirname(filepath)
            target_dir = os.path.join(parent, f"{name_no_ext}_sheets")

            # Si le dossier de sortie existe déjà, on récupère son contenu pour la sortie
            # sans retraiter le fichier source.
            if self.is_excel_already_processed(filepath):
                if os.path.exists(target_dir):
                    for f in os.listdir(target_dir):
                        if f.lower().endswith(".xlsx"):
                            all_files.append([os.path.join(target_dir, f).replace("\\", "/")])
                progress_callback(idx / total * 100)
                continue

            # Traitement réel du fichier
            os.makedirs(target_dir, exist_ok=True)

            try:
                # Mode read_only=True pour économiser la RAM sur les gros fichiers
                # data_only=True pour extraire les valeurs et non les formules
                wb_source = openpyxl.load_workbook(filepath, read_only=True, data_only=True)

                for sheet_name in wb_source.sheetnames:
                    safe_name = "".join(c if c.isalnum() or c in (' ', '-', '_') else "_" for c in sheet_name)
                    out_path = os.path.join(target_dir, f"{safe_name}.xlsx")

                    # Création du nouveau classeur pour la feuille individuelle
                    wb_dest = Workbook()
                    ws_dest = wb_dest.active
                    ws_dest.title = sheet_name

                    ws_source = wb_source[sheet_name]

                    has_data = False
                    # On itère sur les lignes de la source
                    for row in ws_source.rows:
                        row_values = [cell.value for cell in row]
                        # On vérifie si la ligne contient au moins une donnée
                        if any(v is not None for v in row_values):
                            ws_dest.append(row_values)
                            has_data = True

                    if has_data:
                        wb_dest.save(out_path)
                        all_files.append([out_path.replace("\\", "/")])

                wb_source.close()

            except Exception as e:
                print(f"Skipping {filepath} due to error: {e}")
                continue

            progress_callback(idx / total * 100)

        return all_files


if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = OWSplitExcelSheets()
    w.show()
    app.exec()


import os
import sys

import Orange.data
from AnyQt.QtWidgets import QApplication, QLabel, QWidget, QScrollArea, QLineEdit, QHBoxLayout, QVBoxLayout, QDoubleSpinBox, QSpinBox
from AnyQt.QtCore import Qt
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.settings import Setting

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import answers_llama
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.llm import answers_llama
    from orangecontrib.AAIT.utils import thread_management
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWConverseLLM(widget.OWWidget):
    name = "Converse LLM"
    description = "Generate a response to a column 'prompt' with a LLM, while keeping the previous interactions in memory"
    icon = "icons/owconversellm.svg"
    category = "AAIT - ALGORITHM"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owconversellm.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owconversellm.ui")
    want_control_area = False
    priority = 1089

    class Inputs:
        data = Input("Data", Orange.data.Table)
        model_path = Input("Model", str, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    n_ctx: str = Setting("32768")
    workflow_id = Setting("")

    max_tokens = Setting(0)
    temperature = Setting(0.4)
    top_p = Setting(0.4)
    top_k = Setting(40)
    repeat_penalty = Setting(1.15)
    k_cache = Setting(0)
    v_cache = Setting(0)

    @Inputs.data
    def set_data(self, in_data):
        if in_data is None:
            self.Outputs.data.send(None)
            return
        self.data = in_data
        if self.autorun:
            self.run()

    @Inputs.model_path
    def set_model_path(self, in_model_path):
        self.model_path = in_model_path
        # Reset everything when changing model
        self.conversation = []
        self.clear_chat_area() # TODO render Qt UI

        # If there is already a model
        if self.model is not None:
            # Interrupt an eventual threaded generation
            if self.thread is not None:
                if self.thread.isRunning():
                    self.thread.safe_quit()
            self.model = None

        # If the link has been cut, do nothing
        if in_model_path is None:
            return
        # Check for the GPU / CPU, load model and start a chat session # TODO ?
        # answers.check_gpu(in_model_path, self)
        self.model = answers_llama.load_model(self.model_path, use_gpu=True, n_ctx=int(self.n_ctx))

        if self.autorun:
            self.run()


    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(750)
        self.setFixedHeight(650)
        uic.loadUi(self.gui, self)
        # Context
        self.lineEdit_n_ctx = self.findChild(QLineEdit, "lineEdit")
        self.lineEdit_n_ctx.setText(str(self.n_ctx))
        self.lineEdit_n_ctx.editingFinished.connect(self.update_n_ctx)
        # Context cache
        self.comboKCache.setCurrentIndex(self.k_cache)
        self.comboVCache.setCurrentIndex(self.v_cache)
        self.comboKCache.currentIndexChanged.connect(self.update_cache)
        self.comboVCache.currentIndexChanged.connect(self.update_cache)
        # Workflow ID for streaming
        self.edit_ID = self.findChild(QLineEdit, 'lineEdit_2')
        self.edit_ID.setText(self.workflow_id)
        self.edit_ID.editingFinished.connect(self.update_workflow_id)
        # SpinBoxes
        self.edit_max_tokens = self.bind_spinbox("boxMaxTokens", self.max_tokens)
        self.edit_temperature = self.bind_spinbox("boxTemperature", self.temperature, is_double=True)
        self.edit_top_p = self.bind_spinbox("boxTopP", self.top_p, is_double=True)
        self.edit_top_k = self.bind_spinbox("boxTopK", self.top_k)
        self.edit_repeat_penalty = self.bind_spinbox("boxRepeatPenalty", self.repeat_penalty, is_double=True)
        # Chat display
        self.scrollArea_display = self.findChild(QScrollArea, "scrollArea")
        self.scrollArea_display.setWidgetResizable(True)
        self.widget_display = self.scrollArea_display.widget()
        self.vLayout_display = QVBoxLayout(self.widget_display)
        self.vLayout_display.addStretch()

        # Data Management
        self.data = None
        self.model = None
        self.model_path = None
        self.conversation = None
        self.current_assistant_card = None

        self.thread = None
        self.autorun = True
        self.use_gpu = False
        self.can_run = True
        self.result = None

        # Custom updates
        self.post_initialized()

    def update_n_ctx(self):
        value = self.lineEdit_n_ctx.text()
        self.n_ctx = value if value.isdigit() else "32768"
        # answers.check_gpu(self.model_path, self) # TODO reload model

    def update_cache(self):
        self.k_cache = self.comboKCache.currentIndex()
        self.v_cache = self.comboVCache.currentIndex()

    def update_workflow_id(self):
        self.workflow_id = self.edit_ID.text()

    def bind_spinbox(self, name, value, is_double=False):
        widget_type = QDoubleSpinBox if is_double else QSpinBox
        box = self.findChild(widget_type, name)
        box.setValue(value)
        box.editingFinished.connect(self.update_parameters)
        return box

    def update_parameters(self):
        self.max_tokens = self.edit_max_tokens.value()
        self.temperature = self.edit_temperature.value()
        self.top_p = self.edit_top_p.value()
        self.top_k = self.edit_top_k.value()
        self.repeat_penalty = self.edit_repeat_penalty.value()
        print("Updated parameters:")
        print(f"  Max tokens     : {self.max_tokens}")
        print(f"  Temperature    : {self.temperature}")
        print(f"  Top P          : {self.top_p}")
        print(f"  Top K          : {self.top_k}")
        print(f"  Repeat penalty : {self.repeat_penalty}")

    def run(self):
        # Clear error & warning
        self.error("")
        self.warning("")

        # If Thread is already running, interrupt it
        if self.thread is not None:
            if self.thread.isRunning():
                self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if self.model is None:
            self.Outputs.data.send(None)
            return

        if not "prompt" in self.data.domain:
            self.error('You need a "prompt" column in your input data.')
            self.Outputs.data.send(None)
            return

        query_parameters = {
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "repeat_penalty": self.repeat_penalty,
            "k_cache": self.k_cache,
            "v_cache": self.v_cache
        }

        # Start progress bar
        self.progressBarInit()

        # Connect and start thread : main function, progress, result and finish
        # --> progress is used in the main function to track progress (with a callback)
        # --> result is used to collect the result from main function
        # --> finish is just an empty signal to indicate that the thread is finished
        self.thread = thread_management.Thread(answers_llama.generate_conversation, self.data, self.model, self.conversation, int(self.n_ctx), query_parameters, self.workflow_id)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, progress) -> None:
        action = progress[0]
        value = progress[1]
        if action == "progressBar":
            self.progressBarSet(value)
        elif action == "user":
            user_card = UserMessageCard(value)
            self.add_message_card(user_card)
        elif action == "assistant":
            if self.current_assistant_card is None:
                assistant_card = AssistantMessageCard(value)
                self.add_message_card(assistant_card)
                self.current_assistant_card = assistant_card
            else:
                current_text = self.current_assistant_card.label.text()
                current_text += value
                self.current_assistant_card.setText(current_text)
        elif action == "warning":
            self.warning(value)
        # QApplication.processEvents()
        # self.scroll_to_bottom()


    def handle_result(self, result):
        try:
            self.result = result[0]
            self.conversation = result[1]
            self.Outputs.data.send(self.result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        print("Generation finished")
        self.progressBarFinished()
        self.current_assistant_card = None

    def onDeleteWidget(self):
        if self.model is not None:
            self.model = None
        super().onDeleteWidget()

    def post_initialized(self):
        pass


    # UI
    def add_message_card(self, message_card):
        self.vLayout_display.insertWidget(self.vLayout_display.count() - 1, message_card)

    # UI
    def scroll_to_bottom(self):
        self.scrollArea_display.verticalScrollBar().setValue(self.scrollArea_display.verticalScrollBar().maximum())

    # UI # TODO à tester
    def clear_chat_area(self):
        """
        Supprime uniquement les cartes de message sans casser le layout.
        """
        for i in reversed(range(self.vLayout_display.count())):
            item = self.vLayout_display.itemAt(i)
            widget = item.widget()

            # On ne supprime que les widgets qui sont des cartes
            if isinstance(widget, (UserMessageCard, AssistantMessageCard)):
                widget.setParent(None)  # Détache du layout
                widget.deleteLater()  # Supprime de la mémoire
            elif widget is None and item.spacerItem():
                # Optionnel : si tu veux garder ton "ressort" du bas,
                # ne fais rien ici.
                pass



# UI - User message card
class UserMessageCard(QWidget):
    def __init__(self, text=""):
        super().__init__()
        self.layout = QHBoxLayout(self)
        self.label = QLabel(text)
        self.label.setWordWrap(True)
        self.label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.label.setStyleSheet("""
            background-color: #5050a0;
            color: white;
            border-radius: 8px;
            padding: 6px;
        """)
        self.layout.addWidget(self.label)
        self.layout.addStretch()  # push message to left
        self.setLayout(self.layout)

    def setText(self, text):
        """Update the card's text dynamically."""
        self.label.setText(text)

# UI - Assistant message card
class AssistantMessageCard(QWidget):
    def __init__(self, text=""):
        super().__init__()
        self.layout = QHBoxLayout(self)
        self.layout.addStretch()  # push message to right
        self.label = QLabel(text)
        self.label.setWordWrap(True)
        self.label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.label.setStyleSheet("""
            background-color: #3c3c3c;
            color: white;
            border-radius: 8px;
            padding: 6px;
        """)
        self.layout.addWidget(self.label)
        self.setLayout(self.layout)

    def setText(self, text):
        """Update the card's text dynamically."""
        self.label.setText(text)






if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWConverseLLM()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys

from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from Orange.data import Table, Domain, ContinuousVariable
from thefuzz import fuzz
from AnyQt.QtWidgets import QApplication
from AnyQt.QtCore import QObject, QThread, pyqtSignal

# Import intelligent selon contexte
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file

# Worker pour exécution asynchrone
class KeywordWorker(QObject):
    finished = pyqtSignal(Table)
    error = pyqtSignal(str)

    def __init__(self, data, keywords, threshold=80):
        super().__init__()
        self.data = data
        self.keywords = keywords
        self.threshold = threshold

    def extract_matched_keywords(self, text, keywords):
        words = text.split(" ")
        matched_keywords = []
        for keyword in keywords:
            best_score = max(fuzz.ratio(word.lower(), keyword.lower()) for word in words)
            if best_score >= self.threshold:
                matched_keywords.append(keyword)
        return matched_keywords

    def run(self):
        try:
            if "Content" not in self.data.domain or "Keywords" not in self.keywords.domain:
                self.error.emit("Missing 'Content' or 'Keywords' column")
                return

            keyword_list = [
                str(row["Keywords"])
                for row in self.keywords
                if str(row["Keywords"]).strip() != ""
            ]

            new_metas_vars = list(self.data.domain.metas)
            if not any(var.name == "Keywords" for var in new_metas_vars):
                new_metas_vars.append(ContinuousVariable("Keywords"))

            new_domain = Domain(
                self.data.domain.attributes,
                self.data.domain.class_vars,
                new_metas_vars
            )

            new_metas = []
            for row in self.data:
                text = str(row["Content"])
                matched_keywords = self.extract_matched_keywords(text, keyword_list)
                score = (
                    sum(fuzz.ratio(word.lower(), kw.lower()) for kw in matched_keywords for word in text.split())
                    / len(keyword_list)
                    if matched_keywords else 0.0
                )
                new_metas.append(list(row.metas) + [score])

            out_data = Table(new_domain, self.data.X, self.data.Y, new_metas)
            self.finished.emit(out_data)

        except Exception as e:
            self.error.emit(str(e))

# Définition du widget principal
@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWKeywords(widget.OWWidget):
    name = "Keywords Detection"
    description = "Give the amount of keywords from in_object in in_data"
    icon = "icons/owkeywords.png"
    category = "AAIT - LLM INTEGRATION"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owkeywords.png"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owkeyword.ui")
    want_control_area = False
    priority = 1050

    class Inputs:
        data = Input("Content", Table)
        keywords = Input("Keywords", Table)

    class Outputs:
        data = Output("Keywords per Content", Table)

    def __init__(self):
        super().__init__()
        self.data = None
        self.keywords = None
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)
        self.autorun = True
        self.thread = None
        self.worker = None

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.autorun and self.keywords:
            self.process()

    @Inputs.keywords
    def set_keywords(self, in_keywords):
        self.keywords = in_keywords
        if self.autorun and self.data:
            self.process()

    def process(self):
        if self.data is None or self.keywords is None:
            self.Outputs.data.send(None)
            return

        self.error("")  # Clear errors
        self.thread = QThread()
        self.worker = KeywordWorker(self.data, self.keywords)
        self.worker.moveToThread(self.thread)

        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.on_worker_finished)
        self.worker.error.connect(self.on_worker_error)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)

        self.thread.start()

    def on_worker_finished(self, out_data):
        print("[INFO] Thread finished. Output ready.")
        self.Outputs.data.send(out_data)

    def on_worker_error(self, message):
        print(f"[ERROR] Worker error: {message}")
        self.error(message)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    ow = OWKeywords()
    ow.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()




import requests
import json

def get_model_lmstudio():
    # Configuration des options de la requête
    url = "http://127.0.0.1:1234/v1/models"
    headers = {
        "Content-Type": "application/json"
    }

    try:
        # Envoi de la requête POST
        response = requests.get(url, headers=headers)

        # Vérification de la réponse
        if response.status_code != 200:
            raise Exception("Erreur dans la requête")

        # Analyse du JSON de la réponse
        return response.json()

    except Exception as e:
        print(f"Erreur : {e}")
        return "Error"


def appel_lmstudio(prompt, model, stream=False, temperature=0.7, max_tokens=4096):
    # Configuration des options de la requête
    url = "http://127.0.0.1:1234/v1/chat/completions"
    headers = {
        "Content-Type": "application/json"
    }
    data = {
        "model": model,
        "messages": [
            {"role": "user",
             "content": prompt}
        ],
        "temperature": temperature,
        "max_tokens": max_tokens,
        "stream": stream
    }

    try:
        # Envoi de la requête POST
        response = requests.post(url, headers=headers, data=json.dumps(data), stream=True)

        # Vérification de la réponse
        if response.status_code != 200:
            raise Exception("Error in the request")

        # Initialisation de la variable de contenu
        content = ""
        if stream == True:
            # Lecture et traitement du streaming
            for chunk in response.iter_lines(decode_unicode=True):
                if chunk:
                    line = chunk.strip()
                    print(line)
                    if line == "data: [DONE]":
                        break

                    if line.startswith("data: "):
                        json_data = json.loads(line[6:])
                        text = json_data.get("choices", [{}])[0].get("delta", {}).get("content", "")

                        if text:
                            content += text

        if stream == False:
            # Analyse du JSON de la réponse
            response_data = response.json()
            # Extraction du contenu du message
            content = response_data.get("choices", [{}])[0].get("message", {}).get("content", None)
        return content

    except Exception as e:
        print(f"Error : {e}")
        return None




import os
import sys

import Orange.data
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from Orange.widgets.orangecontrib.AAIT.utils.MetManagement import create_trigger_table
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from orangecontrib.AAIT.utils.MetManagement import create_trigger_table


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWEmptySwitch(widget.OWWidget):
    name = "Empty Switch"
    description = "This widget lets the input data pass if it receives any, otherwise it emits a table on its Trigger output."
    category = "AAIT - TOOLBOX"
    icon = "icons/owemptyswitch.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owemptyswitch.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owemptyswitch.ui")
    want_control_area = False
    priority = 1060

    class Inputs:
        data = Input("Data", object, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)
        trigger = Output("Trigger", object, auto_summary=False)


    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.autorun:
            self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        # Data Management
        self.data = None
        self.autorun = True
        self.post_initialized()

    def run(self):
        if self.data is None:
            self.Outputs.data.send(None)
            self.Outputs.trigger.send(create_trigger_table())
        else:
            self.Outputs.data.send(self.data)
            self.Outputs.trigger.send(None)

    def post_initialized(self):
        pass


if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWEmptySwitch()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os

prompt_templates = {
    "llama": {
        "system": "<|begin_of_text|><|start_header_id|>system<|end_header_id|>\n{system_prompt}<|eot_id|>",
        "user": "<|start_header_id|>user<|end_header_id|>\n{user_prompt}<|eot_id|>",
        "assistant": "<|start_header_id|>assistant<|end_header_id|>\n{assistant_prompt}"
    },

    "mistral": {
        "system": "{system_prompt}\n",
        "user": "<s>[INST] {user_prompt} [/INST]</s>\n",
        "assistant": "{assistant_prompt}"
    },

    "solar": {
        "system": "{system_prompt}\n",
        "user": "### User: {user_prompt}\n",
        "assistant": "### Assistant: {assistant_prompt}"
    },

    "deepseek": {
        "system": "{system_prompt}\n",
        "user": "### Instruction: {user_prompt}\n",
        "assistant": "### Response: {assistant_prompt}"
    },

    "qwen": {
        "system": "<|im_start|>system\n{system_prompt}<|im_end|>\n",
        "user": "<|im_start|>user\n{user_prompt}<|im_end|>\n",
        "assistant": "<|im_start|>assistant\n{assistant_prompt}"
    },

    "gemma": {
        "system": "{system_prompt}\n",
        "user": "<start_of_turn>user\n{user_prompt}<end_of_turn>\n",
        "assistant": "<start_of_turn>model\n{assistant_prompt}\n"
    },

    "granite": {
        "system": "<|system|>\n{system_prompt}\n",
        "user": "<|user|>\n{user_prompt}\n",
        "assistant": "<|assistant|>\n{assistant_prompt}"
    },

    "phi": {
        "system": "<|im_start|>system<|im_sep|>\n{system_prompt}<|im_end|>\n",
        "user": "<|im_start|>user<|im_sep|>\n{user_prompt}<|im_end|>\n",
        "assistant": "<|im_start|>assistant<|im_sep|>\n{assistant_prompt}"
    },

    "default": {
        "system": "{system_prompt}",
        "user": "{user_prompt}",
        "assistant": "{assistant_prompt}"
    }
}



stop_tokens = {
    "llama": "<|eot_id|>",
    "mistral": "</s>",
    "qwen": "<|im_end|>",
    "gemma": "<end_of_turn>",
    "granite": "<|endoftext|>",
    "phi": "<|im_end|>"
}


model_types = {
    "solar-10.7b-instruct-v1.0.Q6_K.gguf": "solar",
    "solar-10.7b-instruct-v1.0-uncensored.Q6_K.gguf": "solar",
    "Mistral-7B-Instruct-v0.3.Q6_K.gguf": "mistral",
    "Qwen2.5.1-Coder-7B-Instruct-Q6_K.gguf": "qwen",
    "qwen2.5-3b-instruct-q4_k_m.gguf": "qwen",
    "deepseek-coder-6.7b-instruct.Q6_K.gguf": "deepseek"
}

model_keywords = ["qwen", "solar", "mistral", "llama", "deepseek", "gemma", "granite", "phi"]



def get_model_type(model_path):
    model_name = os.path.basename(model_path)
    model_type = model_types.get(model_name)
    if model_type is None:
        model_type = next((keyword for keyword in model_keywords if keyword in model_name.lower()), None)
    if model_type is None:
        model_type = "default"
    return model_type


def apply_prompt_template(model_path, user_prompt, assistant_prompt="", system_prompt=""):
    """
    Apply a prompt template based on the given model name and user input.

    Parameters:
        model_path (str): The name of the model used to determine its type.
        user_prompt (str): The user input or request to embed into the prompt.
        assistant_prompt (str, optional): The assistant's beginning of response, if any. Defaults to an empty string.
        system_prompt (str, optional): A system-level instruction or context to include in the prompt. Defaults to an empty string.

    Returns:
        str: The formatted prompt that is ready to be passed to the model.
    """
    # Try to identify the model's type
    model_type = get_model_type(model_path)
    # Retrieve the template
    template = prompt_templates.get(model_type, prompt_templates["default"])  # Default template if none found
    # Apply the template
    prompt = ""
    if system_prompt is not None:
        prompt += template["system"].format(system_prompt=system_prompt)
    prompt += template["user"].format(user_prompt=user_prompt)
    prompt += template["assistant"].format(assistant_prompt=assistant_prompt)
    return prompt


def apply_system_template(model_path, system_prompt):
    model_type = get_model_type(model_path)
    template = prompt_templates.get(model_type, prompt_templates["default"])
    return template["system"].format(system_prompt=system_prompt)

def apply_user_template(model_path, user_prompt):
    model_type = get_model_type(model_path)
    template = prompt_templates.get(model_type, prompt_templates["default"])
    return template["user"].format(user_prompt=user_prompt)

def apply_assistant_template(model_path, assistant_prompt):
    model_type = get_model_type(model_path)
    template = prompt_templates.get(model_type, prompt_templates["default"])
    return template["assistant"].format(assistant_prompt=assistant_prompt)


def get_stop_token(model_name):
    """
    Get the stop token according to the model name / type.

    Parameters:
        model_name (str): The name of the model used to determine its type.
    """
    # If there is a stop token
    try:
        # Get the model type
        model_type = model_types[model_name]
        # Get the template for the model type
        stop_token = stop_tokens[model_type]
    except KeyError as e:
        print(f"Your model {model_name} has no stop token defined. See prompt_management.py. (detail: {e})")
        return None
    return stop_token








# mini_notepadpp_tabs.py
# PyQt6 mini Notepad++-like editor (single file)
#
# Features (kept):
# - Tabs (multi-file)
# - Line numbers
# - Syntax highlighting: .txt .py .bat .html/.htm .json .xml .md
# - Open / Save / Save As (Ctrl+O / Ctrl+S / Ctrl+Shift+S)
# - Copy/Cut/Paste (Ctrl+C/X/V)
# - Find (Ctrl+F) / Replace (Ctrl+H)
# - Auto-indentation (Python ':' adds 4 spaces)
# - Comment/Uncomment selection (Ctrl+/) with correct style per extension:
#   .py -> '#', .bat -> 'REM ', .html/.xml/.md -> <!-- ... --> (wrap), .json -> refused
# - Show/Hide special characters (spaces/tabs + EOL markers) (Ctrl+Shift+8)
#   EOL markers differ for LF vs CRLF: LF shows 'LF', CRLF shows 'CR LF' (grey) like Notepad++
# - Distinguish and preserve EOL style per file (LF vs CRLF), displayed in status bar
# - Convert EOL: menu "Fins de ligne" -> LF / CRLF / Toggle (CRLF/LF)
# - Drag & drop file(s) onto the window to open in tab (fix: editor doesn't accept drops)
# - No '*' anywhere in tab titles

import json
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from AnyQt.QtCore import Qt, QRect, QSize, QRegularExpression
from AnyQt.QtGui import (
    QAction,
    QColor,
    QFont,
    QKeySequence,
    QPainter,
    QTextCharFormat,
    QSyntaxHighlighter,
    QTextCursor,
    QTextDocument,
    QTextOption,
)
from AnyQt.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QPlainTextEdit,
    QStatusBar,
    QTabWidget,
    QToolBar,
    QVBoxLayout,
    QWidget,
)


def _fmt(color: str, bold: bool = False, italic: bool = False) -> QTextCharFormat:
    f = QTextCharFormat()
    f.setForeground(QColor(color))
    if bold:
        f.setFontWeight(QFont.Weight.Bold)
    if italic:
        f.setFontItalic(True)
    return f


@dataclass
class Rule:
    pattern: QRegularExpression
    fmt: QTextCharFormat


class SimpleHighlighter(QSyntaxHighlighter):
    """Lightweight syntax highlighter (heuristic)."""

    def __init__(self, doc: QTextDocument):
        super().__init__(doc)
        self._rules: list[Rule] = []
        self.set_language("text")

    def set_language(self, language: str) -> None:
        self._rules = self._build_rules(language)
        self.rehighlight()

    def _build_rules(self, language: str) -> list[Rule]:
        rules: list[Rule] = []

        str_fmt = _fmt("#a31515")
        num_fmt = _fmt("#098658")
        kw_fmt = _fmt("#0000ff", bold=True)
        com_fmt = _fmt("#008000", italic=True)
        tag_fmt = _fmt("#800000", bold=True)
        attr_fmt = _fmt("#ff0000")
        builtin_fmt = _fmt("#795e26", bold=True)
        json_key_fmt = _fmt("#0451a5", bold=True)
        md_heading = _fmt("#0451a5", bold=True)
        md_code = _fmt("#a31515")
        md_quote = _fmt("#008000", italic=True)

        def add(rx: str, f: QTextCharFormat):
            rules.append(Rule(QRegularExpression(rx), f))

        if language == "python":
            add(r"#.*$", com_fmt)
            add(r"(\'\'\'[\s\S]*?\'\'\'|\"\"\"[\s\S]*?\"\"\")", str_fmt)
            add(r"(\'([^\'\\]|\\.)*\'|\"([^\"\\]|\\.)*\")", str_fmt)
            add(r"\b\d+(\.\d+)?([eE][+-]?\d+)?\b", num_fmt)

            py_keywords = [
                "False", "None", "True", "and", "as", "assert", "async", "await",
                "break", "class", "continue", "def", "del", "elif", "else", "except",
                "finally", "for", "from", "global", "if", "import", "in", "is",
                "lambda", "nonlocal", "not", "or", "pass", "raise", "return",
                "try", "while", "with", "yield", "match", "case",
            ]
            add(r"\b(" + "|".join(map(re.escape, py_keywords)) + r")\b", kw_fmt)

            py_builtins = [
                "print", "len", "range", "list", "dict", "set", "tuple", "str",
                "int", "float", "bool", "sum", "min", "max", "open", "enumerate",
                "zip", "map", "filter", "any", "all", "isinstance", "type",
                "super", "property",
            ]
            add(r"\b(" + "|".join(map(re.escape, py_builtins)) + r")\b", builtin_fmt)
            add(r"^\s*@\w+(\.\w+)*", _fmt("#aa00aa", bold=True))

        elif language == "json":
            add(r"\"([^\"\\]|\\.)*\"", str_fmt)
            add(r"\b-?\d+(\.\d+)?([eE][+-]?\d+)?\b", num_fmt)
            add(r"\b(true|false|null)\b", kw_fmt)
            add(r"\"([^\"\\]|\\.)*\"\s*(?=:)", json_key_fmt)
            add(r"[{}\[\],:]", _fmt("#000000"))

        elif language in ("html", "xml"):
            add(r"<!--[\s\S]*?-->", com_fmt)
            add(r"</?\b[a-zA-Z][a-zA-Z0-9:-]*\b", tag_fmt)
            add(r"/?>", tag_fmt)
            add(r"\b[a-zA-Z_:][a-zA-Z0-9:._-]*(?=\=)", attr_fmt)
            add(r"(\'([^\'\\]|\\.)*\'|\"([^\"\\]|\\.)*\")", str_fmt)

        elif language == "bat":
            add(r"^\s*(REM\b.*|::.*)$", com_fmt)
            add(r"^\s*:\w+.*$", _fmt("#800080", bold=True))
            add(r"%[^%]+%", _fmt("#0451a5", bold=True))
            bat_kw = [
                "echo", "set", "if", "else", "for", "in", "do", "call", "goto",
                "exit", "pause", "start", "cd", "dir", "copy", "del", "move",
                "mkdir", "rmdir", "pushd", "popd",
            ]
            add(r"\b(" + "|".join(map(re.escape, bat_kw)) + r")\b", kw_fmt)
            add(r"\"([^\"\\]|\\.)*\"", str_fmt)

        elif language == "md":
            add(r"^#{1,6}\s.*$", md_heading)
            add(r"^>\s.*$", md_quote)
            add(r"`[^`]+`", md_code)
            add(r"^\s*```.*$", md_code)
            add(r"\*\*[^*]+\*\*", _fmt("#795e26", bold=True))
            add(r"\*[^*]+\*", _fmt("#795e26", italic=True))

        return rules

    def highlightBlock(self, text: str) -> None:
        for rule in self._rules:
            it = rule.pattern.globalMatch(text)
            while it.hasNext():
                m = it.next()
                self.setFormat(m.capturedStart(), m.capturedLength(), rule.fmt)


class FindDialog(QDialog):
    def __init__(self, parent: QWidget, editor: QPlainTextEdit):
        super().__init__(parent)
        self.setWindowTitle("Rechercher (Ctrl+F)")
        self.editor = editor

        self.find_input = QLineEdit()
        self.case_cb = QCheckBox("Respecter la casse")
        self.word_cb = QCheckBox("Mot entier")
        self.regex_cb = QCheckBox("Regex")

        btn_next = QPushButton("Suivant")
        btn_prev = QPushButton("Précédent")
        btn_close = QPushButton("Fermer")

        btn_next.clicked.connect(lambda: self._find(forward=True))
        btn_prev.clicked.connect(lambda: self._find(forward=False))
        btn_close.clicked.connect(self.close)

        layout = QVBoxLayout()
        layout.addWidget(QLabel("Rechercher :"))
        layout.addWidget(self.find_input)

        opts = QHBoxLayout()
        opts.addWidget(self.case_cb)
        opts.addWidget(self.word_cb)
        opts.addWidget(self.regex_cb)
        layout.addLayout(opts)

        btns = QHBoxLayout()
        btns.addWidget(btn_next)
        btns.addWidget(btn_prev)
        btns.addStretch(1)
        btns.addWidget(btn_close)
        layout.addLayout(btns)

        self.setLayout(layout)
        self.find_input.returnPressed.connect(lambda: self._find(True))

    def _find(self, forward: bool) -> None:
        needle = self.find_input.text()
        if not needle:
            return

        flags = QTextDocument.FindFlag(0)
        if self.case_cb.isChecked():
            flags |= QTextDocument.FindFlag.FindCaseSensitively
        if self.word_cb.isChecked():
            flags |= QTextDocument.FindFlag.FindWholeWords
        if not forward:
            flags |= QTextDocument.FindFlag.FindBackward

        doc = self.editor.document()
        cur = self.editor.textCursor()

        found = doc.find(QRegularExpression(needle), cur, flags) if self.regex_cb.isChecked() else doc.find(needle, cur, flags)

        if found.isNull():
            # wrap-around
            cur.movePosition(cur.MoveOperation.End if not forward else cur.MoveOperation.Start)
            self.editor.setTextCursor(cur)
            found = doc.find(QRegularExpression(needle), cur, flags) if self.regex_cb.isChecked() else doc.find(needle, cur, flags)

        if not found.isNull():
            self.editor.setTextCursor(found)


class ReplaceDialog(QDialog):
    def __init__(self, parent: QWidget, editor: QPlainTextEdit):
        super().__init__(parent)
        self.setWindowTitle("Rechercher et remplacer (Ctrl+H)")
        self.editor = editor

        self.find_input = QLineEdit()
        self.replace_input = QLineEdit()
        self.case_cb = QCheckBox("Respecter la casse")
        self.word_cb = QCheckBox("Mot entier")
        self.regex_cb = QCheckBox("Regex")

        btn_find = QPushButton("Suivant")
        btn_replace = QPushButton("Remplacer")
        btn_replace_all = QPushButton("Tout remplacer")
        btn_close = QPushButton("Fermer")

        btn_find.clicked.connect(self._find_next)
        btn_replace.clicked.connect(self._replace_one)
        btn_replace_all.clicked.connect(self._replace_all)
        btn_close.clicked.connect(self.close)

        layout = QVBoxLayout()
        layout.addWidget(QLabel("Rechercher :"))
        layout.addWidget(self.find_input)
        layout.addWidget(QLabel("Remplacer par :"))
        layout.addWidget(self.replace_input)

        opts = QHBoxLayout()
        opts.addWidget(self.case_cb)
        opts.addWidget(self.word_cb)
        opts.addWidget(self.regex_cb)
        layout.addLayout(opts)

        btns = QHBoxLayout()
        btns.addWidget(btn_find)
        btns.addWidget(btn_replace)
        btns.addWidget(btn_replace_all)
        btns.addStretch(1)
        btns.addWidget(btn_close)
        layout.addLayout(btns)

        self.setLayout(layout)
        self.find_input.returnPressed.connect(self._find_next)

    def _flags(self) -> QTextDocument.FindFlag:
        flags = QTextDocument.FindFlag(0)
        if self.case_cb.isChecked():
            flags |= QTextDocument.FindFlag.FindCaseSensitively
        if self.word_cb.isChecked():
            flags |= QTextDocument.FindFlag.FindWholeWords
        return flags

    def _find_next(self) -> None:
        needle = self.find_input.text()
        if not needle:
            return
        doc = self.editor.document()
        cur = self.editor.textCursor()
        flags = self._flags()
        found = doc.find(QRegularExpression(needle), cur, flags) if self.regex_cb.isChecked() else doc.find(needle, cur, flags)
        if found.isNull():
            cur.movePosition(cur.MoveOperation.Start)
            self.editor.setTextCursor(cur)
            found = doc.find(QRegularExpression(needle), cur, flags) if self.regex_cb.isChecked() else doc.find(needle, cur, flags)
        if not found.isNull():
            self.editor.setTextCursor(found)

    def _replace_one(self) -> None:
        cur = self.editor.textCursor()
        if not cur.hasSelection():
            self._find_next()
            cur = self.editor.textCursor()
            if not cur.hasSelection():
                return
        cur.insertText(self.replace_input.text())
        self.editor.setTextCursor(cur)
        self._find_next()

    def _replace_all(self) -> None:
        needle = self.find_input.text()
        if not needle:
            return
        repl = self.replace_input.text()

        text = self.editor.toPlainText()
        pattern = needle if self.regex_cb.isChecked() else re.escape(needle)

        flags = 0
        if not self.case_cb.isChecked():
            flags |= re.IGNORECASE
        if self.word_cb.isChecked():
            pattern = r"\b" + pattern + r"\b"

        try:
            new_text, n = re.subn(pattern, repl, text, flags=flags)
        except re.error as e:
            QMessageBox.warning(self, "Regex invalide", str(e))
            return

        self.editor.setPlainText(new_text)
        QMessageBox.information(self, "Remplacement", f"{n} occurrence(s) remplacée(s).")


class LineNumberArea(QWidget):
    def __init__(self, editor: "CodeEditor"):
        super().__init__(editor)
        self.code_editor = editor

    def sizeHint(self) -> QSize:
        return QSize(self.code_editor.line_number_area_width(), 0)

    def paintEvent(self, event):
        self.code_editor.line_number_area_paint_event(event)


class CodeEditor(QPlainTextEdit):
    def __init__(self):
        super().__init__()
        self.setAcceptDrops(False)  # IMPORTANT: avoid dropping filenames into editor

        self._path: Optional[Path] = None
        self._language = "text"
        self._highlighter = SimpleHighlighter(self.document())

        # EOL mode per tab
        self.eol_mode = "LF"  # "LF" or "CRLF"

        # Special chars toggle (spaces/tabs + our EOL markers)
        self._show_special = False

        self.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)
        self.setTabStopDistance(4 * self.fontMetrics().horizontalAdvance(" "))

        font = QFont("Consolas")
        font.setStyleHint(QFont.StyleHint.Monospace)
        font.setPointSize(11)
        self.setFont(font)

        # Line numbers
        self.line_number_area = LineNumberArea(self)
        self.blockCountChanged.connect(self.update_line_number_area_width)
        self.updateRequest.connect(self.update_line_number_area)
        self.update_line_number_area_width(0)

        self.apply_show_special_chars(False)

    # ---------- Path / language ----------
    def path(self) -> Optional[Path]:
        return self._path

    def set_path(self, p: Optional[Path]) -> None:
        self._path = p
        self.apply_language("auto")

    def detect_language(self) -> str:
        if not self._path:
            return "text"
        ext = self._path.suffix.lower()
        return {
            ".py": "python",
            ".json": "json",
            ".html": "html",
            ".htm": "html",
            ".xml": "xml",
            ".bat": "bat",
            ".md": "md",
            ".txt": "text",
        }.get(ext, "text")

    def apply_language(self, language: str) -> None:
        if language == "auto":
            language = self.detect_language()
        self._language = language
        self._highlighter.set_language(language)

    def language(self) -> str:
        return self._language

    # ---------- EOL detection / preservation ----------
    def set_text_from_file_bytes(self, raw_bytes: bytes) -> None:
        try:
            s = raw_bytes.decode("utf-8")
        except UnicodeDecodeError:
            s = raw_bytes.decode("latin-1", errors="replace")

        # ✅ Détection robuste (dominante) LF vs CRLF, et support des fichiers mixtes
        crlf_count = s.count("\r\n")
        # Compter les '\n' qui ne font PAS partie d'un '\r\n'
        lf_count = s.count("\n") - crlf_count
        # (optionnel) anciens Mac: lignes séparées par '\r' seuls
        cr_only_count = s.count("\r") - crlf_count

        if crlf_count > lf_count and crlf_count > 0:
            self.eol_mode = "CRLF"
        elif lf_count > 0:
            self.eol_mode = "LF"
        elif cr_only_count > 0:
            # Rare mais on évite de mentir: on ramène en LF côté interne
            self.eol_mode = "LF"
            s = s.replace("\r", "\n")
        else:
            # Fichier 1 ligne sans fin de ligne -> garder défaut
            # (ne surtout pas forcer CRLF)
            pass

        # Normalisation interne: tout en '\n' pour l'éditeur
        s = s.replace("\r\n", "\n")
        self.setPlainText(s)

    def text_for_saving(self) -> str:
        s = self.toPlainText()

        if self.eol_mode == "CRLF":
            # SIMPLE: enlever TOUS les '\r' puis convertir les '\n' en '\r\n'
            s = s.replace("\r", "")
            s = s.replace("\n", "\r\n")
            return s

        # LF: enlever TOUS les '\r' (au cas où) et garder '\n'
        return s.replace("\r", "")
    # ---------- Special chars display ----------
    def apply_show_special_chars(self, enabled: bool) -> None:
        """Show tabs/spaces via QTextOption + draw EOL markers ourselves."""
        self._show_special = enabled

        opt = self.document().defaultTextOption()
        flags = opt.flags()
        flags &= ~QTextOption.Flag.ShowTabsAndSpaces
        flags &= ~QTextOption.Flag.ShowLineAndParagraphSeparators
        flags &= ~QTextOption.Flag.ShowDocumentTerminator

        if enabled:
            flags |= QTextOption.Flag.ShowTabsAndSpaces

        opt.setFlags(flags)
        self.document().setDefaultTextOption(opt)
        self.viewport().update()

    # ---------- Line numbers ----------
    def line_number_area_width(self) -> int:
        digits = len(str(max(1, self.blockCount())))
        return 10 + self.fontMetrics().horizontalAdvance("9") * digits

    def update_line_number_area_width(self, _):
        self.setViewportMargins(self.line_number_area_width(), 0, 0, 0)

    def update_line_number_area(self, rect, dy):
        if dy:
            self.line_number_area.scroll(0, dy)
        else:
            self.line_number_area.update(0, rect.y(), self.line_number_area.width(), rect.height())
        if rect.contains(self.viewport().rect()):
            self.update_line_number_area_width(0)
        if self._show_special:
            self.viewport().update()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        cr = self.contentsRect()
        self.line_number_area.setGeometry(QRect(cr.left(), cr.top(), self.line_number_area_width(), cr.height()))

    def line_number_area_paint_event(self, event):
        painter = QPainter(self.line_number_area)
        painter.fillRect(event.rect(), QColor("#f3f3f3"))

        block = self.firstVisibleBlock()
        block_number = block.blockNumber()
        top = int(self.blockBoundingGeometry(block).translated(self.contentOffset()).top())
        bottom = top + int(self.blockBoundingRect(block).height())

        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                painter.setPen(QColor("#666666"))
                painter.drawText(
                    0,
                    top,
                    self.line_number_area.width() - 4,
                    self.fontMetrics().height(),
                    Qt.AlignmentFlag.AlignRight,
                    str(block_number + 1),
                )

            block = block.next()
            top = bottom
            bottom = top + int(self.blockBoundingRect(block).height())
            block_number += 1

    # ---------- Auto-indentation ----------
    def keyPressEvent(self, event):
        # --- Tab / Shift+Tab : indent/outdent la sélection (ou la ligne) ---
        if event.key() == Qt.Key.Key_Backtab:
            self.outdent_selection()
            return

        if event.key() == Qt.Key.Key_Tab and not (event.modifiers() & Qt.KeyboardModifier.ControlModifier):
            if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                self.outdent_selection()
            else:
                self.indent_selection()
            return
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            cur = self.textCursor()
            cur.beginEditBlock()

            block_text = cur.block().text()
            indent = re.match(r"\s*", block_text).group(0)

            extra = ""
            if self._language == "python":
                stripped = block_text.strip()
                if stripped.endswith(":") and not stripped.startswith("#"):
                    extra = " " * 4

            super().keyPressEvent(event)
            cur = self.textCursor()
            cur.insertText(indent + extra)
            cur.endEditBlock()
            return

        super().keyPressEvent(event)



    # ---------- EOL markers painting ----------
    def paintEvent(self, event):
        super().paintEvent(event)

        if not self._show_special:
            return

        # Never crash the editor due to marker drawing
        try:
            painter = QPainter(self.viewport())
            painter.setPen(QColor("#9a9a9a"))
            fm = self.fontMetrics()
            tokens = ["LF"] if self.eol_mode == "LF" else ["CR", "LF"]

            block = self.firstVisibleBlock()
            while block.isValid():
                geom = self.blockBoundingGeometry(block).translated(self.contentOffset())
                top = int(geom.top())
                if top > self.viewport().height():
                    break

                if block.isVisible():
                    tc = QTextCursor(block)
                    tc.movePosition(QTextCursor.MoveOperation.EndOfBlock)
                    r = self.cursorRect(tc)
                    x = r.right() + 2
                    y = r.top() + fm.ascent()

                    if -fm.height() <= y <= self.viewport().height() + fm.height():
                        for tok in tokens:
                            painter.drawText(x, y, tok)
                            x += fm.horizontalAdvance(tok) + fm.horizontalAdvance(" ")
                block = block.next()
        except Exception:
            return

    # ==========================================================
    # ADD: Indent / Outdent selection (per-file indent unit)
    # ==========================================================
    def _indent_unit(self) -> str:
        """
        Choisit l'unité d'indentation pour CE document.
        1) Détecte tabs/espaces en scannant les lignes (dominant)
        2) Sinon fallback par langage/extension gérés par ton code
        """
        txt = self.toPlainText()
        tab_hits = 0
        space_indents = []

        for line in txt.splitlines()[:5000]:
            if not line:
                continue
            if line.startswith("\t"):
                tab_hits += 1
                continue
            if line.startswith(" "):
                n = len(line) - len(line.lstrip(" "))
                if n > 0:
                    space_indents.append(n)

        # Si tabs visibles et majoritaires → tabs
        if tab_hits > 0 and tab_hits >= len(space_indents):
            return "\t"

        # Sinon, espaces : estimer le pas, sinon fallback par langage
        step = 0
        if space_indents:
            m = min(space_indents)
            if m in (2, 4, 8):
                step = m
            else:
                step = 4 if m >= 4 else 2

        if step <= 0:
            lang = self.detect_language()
            # Fallbacks raisonnables par format (basés sur conventions courantes)
            # - python: 4
            # - json/html/xml: 2 (souvent)
            # - md: 4 (souvent) ; txt: 4 ; bat: 4
            fallback = {
                "python": 4,
                "json": 2,
                "html": 2,
                "xml": 2,
                "md": 4,
                "bat": 4,
                "text": 4,
            }
            step = fallback.get(lang, 4)

        return " " * step

    def _selected_block_range(self, cursor: QTextCursor):
        """
        Retourne (start_block_number, end_block_number) couvrant la sélection.
        Si la sélection finit pile au début d'une ligne, on n'inclut pas cette ligne.
        """
        doc = self.document()
        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        c1 = QTextCursor(doc)
        c1.setPosition(start)
        start_block = c1.blockNumber()

        c2 = QTextCursor(doc)
        c2.setPosition(end)
        if end > start and c2.atBlockStart():
            c2.movePosition(QTextCursor.MoveOperation.PreviousBlock)
        end_block = c2.blockNumber()

        return start_block, end_block

    def indent_selection(self) -> None:
        cur = self.textCursor()
        if not cur.hasSelection():
            cur.select(QTextCursor.SelectionType.LineUnderCursor)

        unit = self._indent_unit()
        doc = self.document()
        start_block, end_block = self._selected_block_range(cur)

        cur.beginEditBlock()
        for block_no in range(start_block, end_block + 1):
            block = doc.findBlockByNumber(block_no)
            if not block.isValid():
                continue
            c = QTextCursor(block)
            c.movePosition(QTextCursor.MoveOperation.StartOfBlock)
            c.insertText(unit)
        cur.endEditBlock()

    def outdent_selection(self) -> None:
        cur = self.textCursor()
        if not cur.hasSelection():
            cur.select(QTextCursor.SelectionType.LineUnderCursor)

        unit = self._indent_unit()
        is_tabs = (unit == "\t")
        unit_len = len(unit)

        doc = self.document()
        start_block, end_block = self._selected_block_range(cur)

        cur.beginEditBlock()
        for block_no in range(start_block, end_block + 1):
            block = doc.findBlockByNumber(block_no)
            if not block.isValid():
                continue

            text = block.text()
            if not text:
                continue

            c = QTextCursor(block)
            c.movePosition(QTextCursor.MoveOperation.StartOfBlock)

            # cas exact: unité détectée
            if text.startswith(unit):
                c.movePosition(QTextCursor.MoveOperation.Right, QTextCursor.MoveMode.KeepAnchor, unit_len)
                c.removeSelectedText()
                continue

            # si mode espaces: retirer jusqu'à unit_len espaces
            if not is_tabs:
                n = 0
                while n < min(unit_len, len(text)) and text[n] == " ":
                    n += 1
                if n > 0:
                    c.movePosition(QTextCursor.MoveOperation.Right, QTextCursor.MoveMode.KeepAnchor, n)
                    c.removeSelectedText()
                continue

            # mode tabs mais la ligne commence par espaces → ne rien faire (évite casser hybride)
        cur.endEditBlock()


class EditorMainWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("TigerODM — Notepad")
        self.resize(1200, 800)

        self.tabs = QTabWidget()
        self.tabs.setTabsClosable(True)
        self.tabs.tabCloseRequested.connect(self.close_tab)
        self.tabs.currentChanged.connect(self._on_tab_changed)
        self.tabs.setAcceptDrops(True)
        self.setCentralWidget(self.tabs)

        self.setAcceptDrops(True)  # allow dropping onto window

        self.status = QStatusBar()
        self.setStatusBar(self.status)

        self.find_dlg: Optional[FindDialog] = None
        self.replace_dlg: Optional[ReplaceDialog] = None

        self._show_special_chars = False

        self._make_actions()
        self._make_menus()
        self._make_toolbar()

        self.new_tab()

    def current_editor(self) -> Optional[CodeEditor]:
        w = self.tabs.currentWidget()
        return w if isinstance(w, CodeEditor) else None

    def _tab_title(self, ed: CodeEditor) -> str:
        return ed.path().name if ed.path() else "Sans titre"

    def _refresh_tab_titles(self):
        for i in range(self.tabs.count()):
            ed = self.tabs.widget(i)
            if isinstance(ed, CodeEditor):
                self.tabs.setTabText(i, self._tab_title(ed))

    def _update_status(self):
        ed = self.current_editor()
        if not ed:
            self.status.clearMessage()
            return
        name = ed.path().name if ed.path() else "Sans titre"
        self.status.showMessage(f"{name}  |  EOL: {ed.eol_mode}")

    # ---------- UI ----------
    def _make_actions(self):
        self.act_new = QAction("Nouveau", self)
        self.act_new.setShortcut(QKeySequence.StandardKey.New)
        self.act_new.triggered.connect(self.new_tab)

        self.act_open = QAction("Ouvrir…", self)
        self.act_open.setShortcut(QKeySequence.StandardKey.Open)
        self.act_open.triggered.connect(self.open_file)

        self.act_save = QAction("Enregistrer", self)
        self.act_save.setShortcut(QKeySequence.StandardKey.Save)
        self.act_save.triggered.connect(self.save_file)

        self.act_save_as = QAction("Enregistrer sous…", self)
        self.act_save_as.setShortcut(QKeySequence.StandardKey.SaveAs)
        self.act_save_as.triggered.connect(self.save_file_as)

        self.act_close_tab = QAction("Fermer l’onglet", self)
        self.act_close_tab.setShortcut(QKeySequence.StandardKey.Close)
        self.act_close_tab.triggered.connect(lambda: self.close_tab(self.tabs.currentIndex()))

        self.act_quit = QAction("Quitter", self)
        self.act_quit.setShortcut(QKeySequence.StandardKey.Quit)
        self.act_quit.triggered.connect(self.close)

        self.act_cut = QAction("Couper", self)
        self.act_cut.setShortcut(QKeySequence.StandardKey.Cut)
        self.act_cut.triggered.connect(lambda: self.current_editor().cut() if self.current_editor() else None)

        self.act_copy = QAction("Copier", self)
        self.act_copy.setShortcut(QKeySequence.StandardKey.Copy)
        self.act_copy.triggered.connect(lambda: self.current_editor().copy() if self.current_editor() else None)

        self.act_paste = QAction("Coller", self)
        self.act_paste.setShortcut(QKeySequence.StandardKey.Paste)
        self.act_paste.triggered.connect(lambda: self.current_editor().paste() if self.current_editor() else None)

        self.act_find = QAction("Rechercher…", self)
        self.act_find.setShortcut(QKeySequence.StandardKey.Find)
        self.act_find.triggered.connect(self.show_find)

        self.act_replace = QAction("Rechercher et remplacer…", self)
        self.act_replace.setShortcut(QKeySequence.StandardKey.Replace)
        self.act_replace.triggered.connect(self.show_replace)

        self.act_toggle_comment = QAction("Commenter/Décommenter", self)
        self.act_toggle_comment.setShortcut(QKeySequence("Ctrl+/"))
        self.act_toggle_comment.triggered.connect(self.toggle_comment_selection)

        self.act_toggle_special = QAction("Afficher les caractères spéciaux", self)
        self.act_toggle_special.setCheckable(True)
        self.act_toggle_special.setChecked(False)
        self.act_toggle_special.setShortcut(QKeySequence("Ctrl+Shift+8"))
        self.act_toggle_special.triggered.connect(self.toggle_special_chars)

        # EOL conversion
        self.act_eol_lf = QAction("Convertir fins de ligne → LF (\\n)", self)
        self.act_eol_lf.triggered.connect(lambda: self.convert_eol("LF"))

        self.act_eol_crlf = QAction("Convertir fins de ligne → CRLF (\\r\\n)", self)
        self.act_eol_crlf.triggered.connect(lambda: self.convert_eol("CRLF"))

        self.act_eol_toggle = QAction("Convertir fins de ligne → CRLF/LF (basculer)", self)
        self.act_eol_toggle.triggered.connect(lambda: self.convert_eol("TOGGLE"))

        # Language override (per tab)
        self.lang_combo = QComboBox()
        self.lang_combo.addItems(["auto", "text", "python", "json", "html", "xml", "md", "bat"])
        self.lang_combo.currentTextChanged.connect(self._on_lang_changed)


        # ----------------------------------------------------------
        self.act_indent = QAction("Indenter", self)
        self.act_indent.setShortcut(QKeySequence("TAB"))
        self.act_indent.triggered.connect(self.indent_selection)

        self.act_outdent = QAction("Désindenter", self)
        self.act_outdent.setShortcut(QKeySequence("SHIFT+TAB"))
        self.act_outdent.triggered.connect(self.outdent_selection)

    def _make_menus(self):
        m_file = self.menuBar().addMenu("Fichier")
        m_file.addAction(self.act_new)
        m_file.addAction(self.act_open)
        m_file.addSeparator()
        m_file.addAction(self.act_save)
        m_file.addAction(self.act_save_as)
        m_file.addSeparator()
        m_file.addAction(self.act_close_tab)
        m_file.addSeparator()
        m_file.addAction(self.act_quit)

        m_edit = self.menuBar().addMenu("Édition")
        m_edit.addAction(self.act_cut)
        m_edit.addAction(self.act_copy)
        m_edit.addAction(self.act_paste)
        m_edit.addSeparator()
        m_edit.addAction(self.act_find)
        m_edit.addAction(self.act_replace)
        m_edit.addSeparator()
        m_edit.addAction(self.act_toggle_comment)

        # ----------------------------------------------------------
        # ADD: Indent/Outdent menu items
        # ----------------------------------------------------------
        m_edit.addSeparator()
        m_edit.addAction(self.act_indent)
        m_edit.addAction(self.act_outdent)

        m_view = self.menuBar().addMenu("Affichage")
        m_view.addAction(self.act_toggle_special)

        m_eol = self.menuBar().addMenu("Fins de ligne")
        m_eol.addAction(self.act_eol_lf)
        m_eol.addAction(self.act_eol_crlf)
        m_eol.addSeparator()
        m_eol.addAction(self.act_eol_toggle)

    def _make_toolbar(self):
        tb = QToolBar("Main")
        self.addToolBar(tb)
        tb.addAction(self.act_new)
        tb.addAction(self.act_open)
        tb.addAction(self.act_save)
        tb.addAction(self.act_save_as)
        tb.addSeparator()
        tb.addAction(self.act_find)
        tb.addAction(self.act_replace)
        tb.addSeparator()
        tb.addAction(self.act_toggle_comment)

        # ----------------------------------------------------------
        # ADD: Indent/Outdent buttons on toolbar
        # ----------------------------------------------------------
        tb.addAction(self.act_indent)
        tb.addAction(self.act_outdent)

        tb.addSeparator()
        tb.addAction(self.act_toggle_special)
        tb.addSeparator()
        tb.addWidget(QLabel("Langage: "))
        tb.addWidget(self.lang_combo)

    # ----------------------------------------------------------
    # ADD: Indent/Outdent handlers (delegate to current editor)
    # ----------------------------------------------------------
    def indent_selection(self):
        ed = self.current_editor()
        if not ed:
            return
        ed.indent_selection()

    def outdent_selection(self):
        ed = self.current_editor()
        if not ed:
            return
        ed.outdent_selection()

    # ---------- Tabs ----------
    def new_tab(self):
        ed = CodeEditor()
        ed.apply_show_special_chars(self._show_special_chars)
        ed.textChanged.connect(self._update_status)
        idx = self.tabs.addTab(ed, self._tab_title(ed))
        self.tabs.setCurrentIndex(idx)
        self._sync_lang_combo()
        self._update_status()

    def close_tab(self, index: int):
        if index < 0 or index >= self.tabs.count():
            return

        ed = self.tabs.widget(index)
        if isinstance(ed, CodeEditor) and ed.document().isModified():
            r = QMessageBox.question(
                self,
                "Modifications non enregistrées",
                f"Enregistrer avant de fermer “{ed.path().name if ed.path() else 'Sans titre'}” ?",
                QMessageBox.StandardButton.Yes
                | QMessageBox.StandardButton.No
                | QMessageBox.StandardButton.Cancel,
            )
            if r == QMessageBox.StandardButton.Cancel:
                return
            if r == QMessageBox.StandardButton.Yes:
                if not self._save_editor(ed, save_as=False):
                    return

        self.tabs.removeTab(index)
        if self.tabs.count() == 0:
            self.new_tab()
        self._update_status()

    def _on_tab_changed(self, _):
        self._sync_lang_combo()
        self._update_status()

    def _sync_lang_combo(self):
        ed = self.current_editor()
        if not ed:
            return
        self.lang_combo.blockSignals(True)
        self.lang_combo.setCurrentText("auto")
        ed.apply_language("auto")
        self.lang_combo.blockSignals(False)

        self.act_toggle_special.blockSignals(True)
        self.act_toggle_special.setChecked(self._show_special_chars)
        self.act_toggle_special.blockSignals(False)

    # ---------- Drag & Drop ----------
    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.acceptProposedAction()
        else:
            event.ignore()

    def dropEvent(self, event):
        if not event.mimeData().hasUrls():
            return
        for url in event.mimeData().urls():
            local = url.toLocalFile()
            if local:
                self.open_file_path(Path(local))
        event.acceptProposedAction()

    # ---------- Special chars ----------
    def toggle_special_chars(self, checked: bool):
        self._show_special_chars = checked
        for i in range(self.tabs.count()):
            ed = self.tabs.widget(i)
            if isinstance(ed, CodeEditor):
                ed.apply_show_special_chars(checked)
        self._update_status()

    # ---------- Language ----------
    def _on_lang_changed(self, choice: str):
        ed = self.current_editor()
        if not ed:
            return
        ed.apply_language(choice)

    # ---------- EOL conversion ----------
    def convert_eol(self, mode: str):
        ed = self.current_editor()
        if not ed:
            return

        if mode == "TOGGLE":
            mode = "LF" if ed.eol_mode == "CRLF" else "CRLF"

        if mode not in ("LF", "CRLF"):
            return

        if ed.eol_mode != mode:
            ed.eol_mode = mode
            ed.document().setModified(True)
            ed.viewport().update()
            self._update_status()
            self.status.showMessage(f"Fins de ligne converties → {mode}. (Sauvegarde requise)", 2500)
        else:
            self.status.showMessage(f"Déjà en {mode}.", 1500)

    # ---------- File IO ----------
    def open_file_path(self, p: Path):
        # focus if already open
        try:
            pr = p.resolve()
        except Exception:
            pr = p

        for i in range(self.tabs.count()):
            ed = self.tabs.widget(i)
            if isinstance(ed, CodeEditor) and ed.path():
                try:
                    if ed.path().resolve() == pr:
                        self.tabs.setCurrentIndex(i)
                        return
                except Exception:
                    pass

        try:
            raw = p.read_bytes()
        except Exception as e:
            QMessageBox.critical(self, "Erreur", str(e))
            return

        ed = CodeEditor()
        ed.apply_show_special_chars(self._show_special_chars)
        ed.set_text_from_file_bytes(raw)
        ed.set_path(p)
        ed.document().setModified(False)
        ed.textChanged.connect(self._update_status)

        idx = self.tabs.addTab(ed, self._tab_title(ed))
        self.tabs.setCurrentIndex(idx)

        ed.apply_language("auto")
        self._refresh_tab_titles()
        self._sync_lang_combo()
        self._update_status()

    def open_file(self):
        filt = "Fichiers supportés (*.txt *.py *.bat *.html *.htm *.json *.xml *.md);;Tous les fichiers (*.*)"
        path_str, _ = QFileDialog.getOpenFileName(self, "Ouvrir un fichier", "", filt)
        if not path_str:
            return
        self.open_file_path(Path(path_str))

    def save_file(self) -> bool:
        ed = self.current_editor()
        if not ed:
            return False
        return self._save_editor(ed, save_as=False)

    def save_file_as(self) -> bool:
        ed = self.current_editor()
        if not ed:
            return False
        return self._save_editor(ed, save_as=True)

    def _save_editor(self, ed: CodeEditor, save_as: bool) -> bool:
        p = ed.path()
        if save_as or p is None:
            filt = "Fichiers supportés (*.txt *.py *.bat *.html *.htm *.json *.xml *.md);;Tous les fichiers (*.*)"
            path_str, _ = QFileDialog.getSaveFileName(self, "Enregistrer sous", "", filt)
            if not path_str:
                return False
            p = Path(path_str)
            ed.set_path(p)
            self._refresh_tab_titles()

        try:
            if p.suffix.lower() == ".json":
                try:
                    json.loads(ed.toPlainText())
                except Exception as e:
                    r = QMessageBox.warning(
                        self,
                        "JSON invalide",
                        f"Le JSON semble invalide:\n{e}\n\nEnregistrer quand même ?",
                        QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    )
                    if r != QMessageBox.StandardButton.Yes:
                        return False

            with open(p, "w", encoding="utf-8", newline="") as f:
                f.write(ed.text_for_saving())
        except Exception as e:
            QMessageBox.critical(self, "Erreur", str(e))
            return False

        ed.document().setModified(False)
        self._update_status()
        self.status.showMessage(f"Enregistré  |  EOL: {ed.eol_mode}", 2000)
        return True

    # ---------- Find / Replace ----------
    def show_find(self):
        ed = self.current_editor()
        if not ed:
            return
        if self.find_dlg is None:
            self.find_dlg = FindDialog(self, ed)
        else:
            self.find_dlg.editor = ed
        self.find_dlg.show()
        self.find_dlg.raise_()
        self.find_dlg.activateWindow()
        self.find_dlg.find_input.setFocus()

    def show_replace(self):
        ed = self.current_editor()
        if not ed:
            return
        if self.replace_dlg is None:
            self.replace_dlg = ReplaceDialog(self, ed)
        else:
            self.replace_dlg.editor = ed
        self.replace_dlg.show()
        self.replace_dlg.raise_()
        self.replace_dlg.activateWindow()
        self.replace_dlg.find_input.setFocus()

    # ---------- Comment / Uncomment ----------
    def toggle_comment_selection(self):
        ed = self.current_editor()
        if not ed:
            return

        lang = ed.language()
        if lang == "text":
            QMessageBox.information(self, "Commentaire", "Aucun style de commentaire pour ce type de fichier.")
            return
        if lang == "json":
            QMessageBox.warning(
                self,
                "Commentaire JSON",
                "JSON standard n'accepte pas les commentaires.\nJe ne commente pas pour éviter de casser le JSON.",
            )
            return

        cur = ed.textCursor()
        if not cur.hasSelection():
            cur.select(QTextCursor.SelectionType.LineUnderCursor)

        start = cur.selectionStart()
        end = cur.selectionEnd()

        doc = ed.document()
        c1 = QTextCursor(doc)
        c1.setPosition(start)
        start_block = c1.block()

        c2 = QTextCursor(doc)
        c2.setPosition(end)
        end_block = c2.block()

        if lang in ("html", "xml", "md"):
            self._toggle_block_comment(ed, start, end, "<!--", "-->")
            ed.document().setModified(True)
            return

        if lang == "python":
            prefix = "#"
        elif lang == "bat":
            prefix = "REM "
        else:
            QMessageBox.information(self, "Commentaire", f"Langage non géré: {lang}")
            return

        blocks = []
        b = start_block
        while b.isValid():
            blocks.append(b)
            if b == end_block:
                break
            b = b.next()

        def is_commented(line: str) -> bool:
            s = line.lstrip()
            return s.startswith(prefix) if prefix != "#" else s.startswith("#")

        non_empty = [bl.text() for bl in blocks if bl.text().strip() != ""]
        do_uncomment = bool(non_empty) and all(is_commented(t) for t in non_empty)

        cur.beginEditBlock()
        for bl in blocks:
            line = bl.text()
            if line.strip() == "":
                continue
            tc = QTextCursor(bl)
            tc.movePosition(QTextCursor.MoveOperation.StartOfBlock)

            if do_uncomment:
                leading = re.match(r"\s*", line).group(0)
                rest = line[len(leading):]
                if rest.startswith(prefix):
                    tc.movePosition(QTextCursor.MoveOperation.Right, QTextCursor.MoveMode.MoveAnchor, len(leading))
                    tc.movePosition(QTextCursor.MoveOperation.Right, QTextCursor.MoveMode.KeepAnchor, len(prefix))
                    tc.removeSelectedText()
                elif prefix == "#" and rest.startswith("#"):
                    tc.movePosition(QTextCursor.MoveOperation.Right, QTextCursor.MoveMode.MoveAnchor, len(leading))
                    tc.movePosition(QTextCursor.MoveOperation.Right, QTextCursor.MoveMode.KeepAnchor, 1)
                    tc.removeSelectedText()
            else:
                leading = re.match(r"\s*", line).group(0)
                tc.movePosition(QTextCursor.MoveOperation.Right, QTextCursor.MoveMode.MoveAnchor, len(leading))
                tc.insertText(prefix)

        cur.endEditBlock()
        ed.setTextCursor(cur)
        ed.document().setModified(True)

    def _toggle_block_comment(self, ed: CodeEditor, start: int, end: int, open_tok: str, close_tok: str):
        doc = ed.document()
        s = min(start, end)
        e = max(start, end)

        text_all = ed.toPlainText()
        left_ws = len(re.match(r"[ \t]*", text_all[s:]).group(0))
        s2 = s + left_ws
        e2 = e

        wrapped = False
        if s2 >= len(open_tok) and text_all[s2 - len(open_tok):s2] == open_tok:
            if text_all[e2:e2 + len(close_tok)] == close_tok:
                wrapped = True

        ed.textCursor().beginEditBlock()
        if wrapped:
            tc2 = QTextCursor(doc)
            tc2.setPosition(e2)
            tc2.setPosition(e2 + len(close_tok), QTextCursor.MoveMode.KeepAnchor)
            tc2.removeSelectedText()

            tc1 = QTextCursor(doc)
            tc1.setPosition(s2 - len(open_tok))
            tc1.setPosition(s2, QTextCursor.MoveMode.KeepAnchor)
            tc1.removeSelectedText()
        else:
            tc3 = QTextCursor(doc)
            tc3.setPosition(e)
            tc3.insertText(close_tok)

            tc4 = QTextCursor(doc)
            tc4.setPosition(s)
            tc4.insertText(open_tok)
        ed.textCursor().endEditBlock()

    # ---------- Close ----------
    def closeEvent(self, event):
        for i in range(self.tabs.count()):
            ed = self.tabs.widget(i)
            if isinstance(ed, CodeEditor) and ed.document().isModified():
                self.tabs.setCurrentIndex(i)
                r = QMessageBox.question(
                    self,
                    "Modifications non enregistrées",
                    f"Enregistrer avant de quitter “{ed.path().name if ed.path() else 'Sans titre'}” ?",
                    QMessageBox.StandardButton.Yes
                    | QMessageBox.StandardButton.No
                    | QMessageBox.StandardButton.Cancel,
                )
                if r == QMessageBox.StandardButton.Cancel:
                    event.ignore()
                    return
                if r == QMessageBox.StandardButton.Yes:
                    if not self._save_editor(ed, save_as=False):
                        event.ignore()
                        return
        event.accept()


def main():
    app = QApplication(sys.argv)
    w = EditorMainWindow()
    w.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



"""Widget Orange pour sélectionner dynamiquement des colonnes.

Ce module fournit un widget qui filtre les colonnes d'une table Orange
en fonction de deux flux d'entrée optionnels:
- un flux de noms de colonnes à faire correspondre de manière stricte;
- un flux de noms utilisés comme motifs « contient » (insensible à la casse).

Le widget renvoie une nouvelle table ne contenant que les variables
dont le nom correspond à l'un ou l'autre des critères.
"""

import sys

import Orange
import Orange.data
import os
from AnyQt.QtWidgets import QApplication
from AnyQt.QtWidgets import (
    QWidget,
    QSizePolicy,
)
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from typing import Optional, Set, Tuple, List

# Import conditionnel pour reproduire l'arborescence des add-ons Orange
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic  # noqa: F401 (kept for consistency)
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic  # type: ignore  # noqa: F401
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file  # type: ignore


def _all_variables(table: Optional[Orange.data.Table]) -> Tuple[List[Orange.data.Variable], List[Orange.data.Variable], List[Orange.data.Variable]]:
    """Retourne les listes (attributs, classes, métas) d'une table.

    - Si la table est ``None``, renvoie trois listes vides.

    Paramètres
    ----------
    table: Orange.data.Table | None
        Table Orange en entrée.

    Retour
    ------
    tuple[list[Variable], list[Variable], list[Variable]]
        Listes des variables d'attributs, de classes et de métadonnées.
    """
    if table is None:
        return [], [], []
    d = table.domain
    return list(d.attributes), list(d.class_vars), list(d.metas)


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWSelectColumnDynamique(widget.OWWidget):
    """Widget pour filtrer des colonnes selon des noms fournis.

    Deux entrées permettent de piloter la sélection:
    - « Columns Strict (names) »: sélection par noms exacts;
    - « Columns Contains (names) »: sélection par inclusion (sans casse).
    """
    name = "Select Columns Dynamic"
    description = "Filter Data columns based on two inputs: exact names and ‘contient’ names."
    category = "AAIT - TOOLBOX"
    icon = "icons/owselectcolumndynamique.png"  # placeholder icon
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owselectcolumndynamique.png"  # dev icon if available
    priority = 1061

    want_main_area = True
    want_control_area = False

    class Inputs:
        data = Input("Data", Orange.data.Table)
        columns_strict = Input("Columns Strict (names)", Orange.data.Table)
        columns_contains = Input("Columns Contains (names)", Orange.data.Table)

    class Outputs:
        data = Output("Filtered Data", Orange.data.Table)

    def __init__(self) -> None:
        """Initialise le widget et charge l'UI (si présente).

        - Prépare les références aux tables d'entrée (données, colonnes strictes,
          colonnes par inclusion).
        - Ajoute la forme au `mainArea`.
        """
        super().__init__()
        self.gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owselect_column_dynamic.ui")

        self._data: Optional[Orange.data.Table] = None
        self._columns_driver_strict: Optional[Orange.data.Table] = None
        self._columns_driver_contains: Optional[Orange.data.Table] = None
        self.form: QWidget = uic.loadUi(self.gui)
        self.mainArea.layout().addWidget(self.form)
        self.form.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        # Pas d'UI interactive: le comportement est entièrement piloté par les données

    # ---------- Inputs ----------
    @Inputs.data
    def set_data(self, data: Optional[Orange.data.Table]) -> None:
        """Réceptionne la table de données principale puis applique le filtre.

        Paramètres
        ----------
        data: Orange.data.Table | None
            Table à filtrer.
        """
        self._data = data
        self._apply()

    @Inputs.columns_strict
    def set_columns_strict(self, data: Optional[Orange.data.Table]) -> None:
        """Réceptionne les noms de colonnes à faire correspondre strictement.

        Attendu: exactement une variable méta `StringVariable` contenant les noms.
        Applique ensuite le filtre.
        """
        self._columns_driver_strict = data
        self._apply()

    @Inputs.columns_contains
    def set_columns_contains(self, data: Optional[Orange.data.Table]) -> None:
        """Réceptionne les motifs de noms de colonnes (recherche par inclusion).

        Attendu: exactement une variable méta `StringVariable` contenant les motifs
        (comparaison insensible à la casse). Applique ensuite le filtre.
        """
        self._columns_driver_contains = data
        self._apply()

    # ---------- Helper ----------

    def _collect_names_from_table(self, df: Optional[Orange.data.Table]) -> Set[str]:
        """Extrait l'ensemble des noms depuis une table de pilotage.

        La table doit contenir exactement une colonne méta de type
        `StringVariable`. Chaque ligne fournit un nom ou un motif.

        Paramètres
        ----------
        df: Orange.data.Table | None
            Table contenant une unique méta chaîne.

        Retour
        ------
        set[str]
            Ensemble des valeurs (chaînes) extraites.
        """
        names: Set[str] = set()
        if df is None:
            return names
        # N'autoriser qu'une unique variable méta chaîne (même politique que SelectRowsDynamic)
        total_columns = len(df.domain.attributes) + len(df.domain.class_vars) + len(df.domain.metas)
        if total_columns != 1:
            return names
        if len(df.domain.metas) != 1:
            return names
        if not isinstance(df.domain.metas[0], Orange.data.StringVariable):
            return names
        for row in df:
            val = row.metas[0]
            if val is None:
                continue
            names.add(str(val))
        return names

    def _sync_selection_from_driver(self) -> Set[str]:
        """Fusionne les noms en provenance des deux entrées.

        Méthode conservée pour compatibilité si appelée à l'extérieur.

        Retour
        ------
        set[str]
            Union des noms stricts et des motifs « contient ».
        """
        s1 = self._collect_names_from_table(self._columns_driver_strict)
        s2 = self._collect_names_from_table(self._columns_driver_contains)
        return s1.union(s2)

    # ---------- Core ----------
    def _apply(self) -> None:
        """Applique le filtrage des colonnes et émet la table filtrée.

        - Récupère les listes de noms stricts et de motifs « contient »;
        - Sélectionne les variables du domaine correspondant aux critères;
        - Construit une nouvelle table si au moins une variable correspond;
        - Émet la sortie « Filtered Data ».
        """
        # S'assurer que des données sont présentes
        data = self._data
        if data is None:
            self.Outputs.data.send(None)
            return

        # Rassembler les noms sélectionnés (depuis les tables de pilotage)
        # Noms stricts et motifs « contient » (comparaison insensible à la casse)
        strict_names = {n.strip() for n in self._collect_names_from_table(self._columns_driver_strict) if n and n.strip()}
        contains_patterns = {n.strip() for n in self._collect_names_from_table(self._columns_driver_contains) if n and n.strip()}
        patterns_l = [n.lower() for n in contains_patterns]

        attrs, class_vars, metas = _all_variables(data)
        def match(vname: str) -> bool:
            if vname in strict_names:
                return True
            vn = vname.lower()
            return any(pat in vn for pat in patterns_l)

        selected_attrs = [v for v in attrs if match(v.name)]
        selected_class = [v for v in class_vars if match(v.name)]
        selected_metas = [v for v in metas if match(v.name)]

        # Construire la table filtrée si nécessaire
        if selected_attrs or selected_class or selected_metas:
            dom_sel = Orange.data.Domain(selected_attrs, selected_class, selected_metas)
            tbl_sel = Orange.data.Table.from_table(dom_sel, data)
        else:
            tbl_sel = None

        self.Outputs.data.send(tbl_sel)

    class Error(widget.OWWidget.Error):
        pass


if __name__ == "__main__":
    # Exécution autonome pour tester le widget manuellement
    app = QApplication(sys.argv)
    w = OWSelectColumnDynamique()
    w.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
from AnyQt.QtWidgets import QPushButton, QApplication, QRadioButton, QComboBox,QCheckBox,QLineEdit
import sys
import numpy as np
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Output, Input
from Orange.data import Table, Domain, StringVariable


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import windows_utils,MetManagement,mac_utils
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
else:
    from orangecontrib.AAIT.utils import windows_utils, MetManagement,mac_utils
    from orangecontrib.AAIT.utils.import_uic import uic


class OWInputSelector(OWWidget):
    name = "Input Selector"
    description = "Select a file or a folder and assign it as a path"
    icon = "icons/in_or_out.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/in_or_out.png"
    category = "AAIT - TOOLBOX"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/ow_in_or_out_path.ui")
    priority = 10
    want_control_area = False
    class Inputs:
        input_table = Input("in data", Table)

    class Outputs:
        path = Output("out data", Table)
        path_string = Output("Data (str)", str, auto_summary=False)


    select_folder_radio_button_checked: str = Setting("True")
    select_multiple_file_checked: str = Setting("False")
    new_file:str = Setting("False")
    combo_box_current_text: str=Setting("All (*.*)")
    line_edit_value:str=Setting("")
    dialog_window_title: str = Setting("")
    str_checkBox_3_send: str=Setting("True")
    str_Launch_automatically_on_startup: str=Setting("False")



    def update_qt_view_from_settings(self):
        if str(self.select_folder_radio_button_checked) =="True":
            self.radioButton.setChecked(True)
            self.radioButton2.setChecked(False)
            self.checkBox.hide()
            self.comboBox.hide()
            self.checkBox_new_file.hide()
            self.update_dialog_title_placeholder()
            return
        self.radioButton.setChecked(False)
        self.radioButton2.setChecked(True)

        if str(self.select_multiple_file_checked)=="False":
            self.checkBox.setChecked(False)
        else:
            self.checkBox.setChecked(True)

        if str(self.new_file)=="False":
            self.checkBox_new_file.setChecked(False)
            self.checkBox.setEnabled(True)
        else:
            self.checkBox_new_file.setChecked(True)
            self.checkBox.setEnabled(False)

        index = self.comboBox.findText(str(self.combo_box_current_text))
        if index != -1:
            self.comboBox.setCurrentIndex(index)
        else:
            self.comboBox.setCurrentIndex(0)  # Sélectionner le premier item




        if self.str_Launch_automatically_on_startup != "False":
            self.checkBox_launch_at_startup.setChecked(True)
        else:
            self.checkBox_launch_at_startup.setChecked(False)





        self.checkBox.show()
        self.comboBox.show()
        self.checkBox_new_file.show()
        self.update_dialog_title_placeholder()


    def update_setting_from_qt_view(self):
        if self.checkBox_launch_at_startup.isChecked():
            self.str_Launch_automatically_on_startup = "True"
        else:
            self.str_Launch_automatically_on_startup ="False"



        if self.radioButton.isChecked():
            self.select_folder_radio_button_checked=True
            self.update_qt_view_from_settings()
            return
        self.select_folder_radio_button_checked = False
        if self.checkBox.isChecked():
            self.select_multiple_file_checked="True"
        else:
            self.select_multiple_file_checked = "False"

        if self.checkBox_new_file.isChecked():
            self.new_file="True"
        else:
            self.new_file = "False"


        self.combo_box_current_text=str(self.comboBox.currentText())
        self.update_qt_view_from_settings()


    # evite boucle infinie
    def on_text_changed(self,new_text):
        if new_text == self.combo_box_current_text:
            return  # Rien à faire, pas de vrai changement
        self.update_setting_from_qt_view()


    def __init__(self):
        super().__init__()
        self.selected_path = ""
        self.in_data = None


        self.setFixedWidth(700)
        self.setFixedHeight(400)
        uic.loadUi(self.gui, self)
        self.radioButton=self.findChild(QRadioButton, 'radioButton')
        self.radioButton2 = self.findChild(QRadioButton, 'radioButton_2')
        self.checkBox = self.findChild(QCheckBox, 'checkBox')
        self.checkBox_new_file = self.findChild(QCheckBox, 'checkBox_2')
        self.comboBox = self.findChild(QComboBox, 'comboBox')
        self.pushButton = self.findChild(QPushButton, 'pushButton')
        self.lineEdit=self.findChild(QLineEdit,'lineEdit')
        self.lineEdit.setText(self.line_edit_value)
        self.windowTitleEdit = self.findChild(QLineEdit, 'lineEdit_window_title')
        self.checkBox_launch_at_startup = self.findChild(QCheckBox, 'checkBox_launch_at_startup')



        if self.str_Launch_automatically_on_startup=="True":
            self.checkBox_launch_at_startup.setChecked(True)
        else:
            self.checkBox_launch_at_startup.setChecked(False)

        if self.windowTitleEdit is not None:
            sanitized_title = (self.dialog_window_title or "").strip()
            if sanitized_title != self.dialog_window_title:
                self.dialog_window_title = sanitized_title
            self.windowTitleEdit.setText(sanitized_title)
            self.update_dialog_title_placeholder()
            self.windowTitleEdit.editingFinished.connect(self.update_dialog_title_setting)

        # based on the string you can modify orders and add without problem
        file_types = [
            "All (*.*)",
            "Images (*.jpg;*.jpeg;*.png;*.bmp;*.gif;*.tiff;*.svg;*.webp;*.heic;*.JPG;*.JPEG;*.PNG;*.BMP;*.GIF;*.TIFF;*.SVG;*.WEBP;*.HEIC)",
            "Documents (*.pdf;*.doc;*.docx;*.xls;*.xlsx;*.ppt;*.pptx;*.txt;*.rtf;*.odt;*.ods;*.odp;*.epub;*.PDF;*.DOC;*.DOCX;*.XLS;*.XLSX;*.PPT;*.PPTX;*.TXT;*.RTF;*.ODT;*.ODS;*.ODP;*.EPUB)",
            "Archives (*.zip;*.rar;*.7z;*.tar;*.gz;*.bz2;*.xz;*.iso;*.ZIP;*.RAR;*.7Z;*.TAR;*.GZ;*.BZ2;*.XZ;*.ISO)",
            "Audio (*.mp3;*.wav;*.ogg;*.flac;*.aac;*.m4a;*.wma;*.aiff;*.alac;*.MP3;*.WAV;*.OGG;*.FLAC;*.AAC;*.M4A;*.WMA;*.AIFF;*.ALAC)",
            "Vidéos (*.mp4;*.avi;*.mkv;*.mov;*.flv;*.wmv;*.mpeg;*.webm;*.3gp;*.MP4;*.AVI;*.MKV;*.MOV;*.FLV;*.WMV;*.MPEG;*.WEBM;*.3GP)",
            "Scripts (*.py;*.pyw;*.js;*.ts;*.html;*.htm;*.css;*.scss;*.sass;*.php;*.rb;*.pl;*.sh;*.c;*.cpp;*.h;*.hpp;*.java;*.swift;*.go;*.kt;*.rs;*.PY;*.PYW;*.JS;*.TS;*.HTML;*.HTM;*.CSS;*.SCSS;*.SASS;*.PHP;*.RB;*.PL;*.SH;*.C;*.CPP;*.H;*.HPP;*.JAVA;*.SWIFT;*.GO;*.KT;*.RS)",
            "Bases de données (*.db;*.sqlite;*.sql;*.csv;*.tsv;*.accdb;*.DB;*.SQLITE;*.SQL;*.CSV;*.TSV;*.ACCDB)",
            "3D/CAO (*.stl;*.obj;*.fbx;*.dae;*.blend;*.3ds;*.dxf;*.dwg;*.gltf;*.GLTF;*.STL;*.OBJ;*.FBX;*.DAE;*.BLEND;*.3DS;*.DXF;*.DWG)",
            "Fonts (*.ttf;*.otf;*.woff;*.woff2;*.eot;*.TTF;*.OTF;*.WOFF;*.WOFF2;*.EOT)",
            "Executables (*.exe;*.msi;*.bat;*.sh;*.apk;*.app;*.EXE;*.MSI;*.BAT;*.SH;*.APK;*.APP)",
            "Images disque (*.iso;*.img;*.dmg;*.ISO;*.IMG;*.DMG)",
            "Données (*.json;*.xml;*.yaml;*.yml;*.ini;*.cfg;*.JSON;*.XML;*.YAML;*.YML;*.INI;*.CFG)",
            "Vecteurs (*.svg;*.eps;*.pdf;*.ai;*.SVG;*.EPS;*.PDF;*.AI)",
            "Notes (*.md;*.markdown;*.MD;*.MARKDOWN)",

            # Fichiers individuels
            "Fichier JPG (*.jpg;*.JPG,*.jpeg;*.JPEG)",
            "Fichier PNG (*.png;*.PNG)",
            "Fichier BMP (*.bmp;*.BMP)",
            "Fichier GIF (*.gif;*.GIF)",
            "Fichier TIFF (*.tiff;*.TIFF)",
            "Fichier SVG (*.svg;*.SVG)",
            "Fichier WEBP (*.webp;*.WEBP)",
            "Fichier HEIC (*.heic;*.HEIC)",
            "Fichier PDF (*.pdf;*.PDF)",
            "Fichier DOC (*.doc;*.DOC)",
            "Fichier DOCX (*.docx;*.DOCX)",
            "Fichier XLS (*.xls;*.XLS)",
            "Fichier XLSX (*.xlsx;*.XLSX)",
            "Fichier PPT (*.ppt;*.PPT)",
            "Fichier PPTX (*.pptx;*.PPTX)",
            "Fichier TXT (*.txt;*.TXT)",
            "Fichier RTF (*.rtf;*.RTF)",
            "Fichier ODT (*.odt;*.ODT)",
            "Fichier ODS (*.ods;*.ODS)",
            "Fichier ODP (*.odp;*.ODP)",
            "Fichier EPUB (*.epub;*.EPUB)",
            "Fichier ZIP (*.zip;*.ZIP)",
            "Fichier RAR (*.rar;*.RAR)",
            "Fichier 7Z (*.7z;*.7Z)",
            "Fichier TAR (*.tar;*.TAR)",
            "Fichier GZ (*.gz;*.GZ)",
            "Fichier BZ2 (*.bz2;*.BZ2)",
            "Fichier XZ (*.xz;*.XZ)",
            "Fichier ISO (*.iso;*.ISO)",
            "Fichier MP3 (*.mp3;*.MP3)",
            "Fichier WAV (*.wav;*.WAV)",
            "Fichier OGG (*.ogg;*.OGG)",
            "Fichier FLAC (*.flac;*.FLAC)",
            "Fichier AAC (*.aac;*.AAC)",
            "Fichier M4A (*.m4a;*.M4A)",
            "Fichier WMA (*.wma;*.WMA)",
            "Fichier AIFF (*.aiff;*.AIFF)",
            "Fichier ALAC (*.alac;*.ALAC)",
            "Fichier MP4 (*.mp4;*.MP4)",
            "Fichier AVI (*.avi;*.AVI)",
            "Fichier MKV (*.mkv;*.MKV)",
            "Fichier MOV (*.mov;*.MOV)",
            "Fichier FLV (*.flv;*.FLV)",
            "Fichier WMV (*.wmv;*.WMV)",
            "Fichier MPEG (*.mpeg;*.MPEG)",
            "Fichier WEBM (*.webm;*.WEBM)",
            "Fichier 3GP (*.3gp;*.3GP)",
            "Fichier PY (*.py;*.PY)",
            "Fichier JS (*.js;*.JS)",
            "Fichier HTML (*.html;*.HTML)",
            "Fichier CSS (*.css;*.CSS)",
            "Fichier PHP (*.php;*.PHP)",
            "Fichier SH (*.sh;*.SH)",
            "Fichier JSON (*.json;*.JSON)",
            "Fichier XML (*.xml;*.XML)",
            "Fichier YAML (*.yaml;*.yml;*.YAML;*.YML)",
            "Fichier EXE (*.exe;*.EXE)",
            "Fichier MSI (*.msi;*.MSI)",
            "Fichier APK (*.apk;*.APK)",
            "Fichier APP (*.app;*.APP)",
            "Fichier DB (*.db;*.DB)",
            "Fichier SQLITE (*.sqlite;*.SQLITE)",
            "Fichier SQL (*.sql;*.SQL)",
            "Fichier CSV (*.csv;*.CSV)",
            "Fichier TSV (*.tsv;*.TSV)",
            "Fichier STL (*.stl;*.STL)",
            "Fichier OBJ (*.obj;*.OBJ)",
            "Fichier FBX (*.fbx;*.FBX)",
            "Fichier DAE (*.dae;*.DAE)",
            "Fichier 3DS (*.3ds;*.3DS)",
            "Fichier DXF (*.dxf;*.DXF)",
            "Fichier DWG (*.dwg;*.DWG)",
            "Fichier GLTF (*.gltf;*.GLTF)",
            "Fichier TTF (*.ttf;*.TTF)",
            "Fichier OTF (*.otf;*.OTF)",
            "Fichier WOFF (*.woff;*.WOFF)",
            "Fichier WOFF2 (*.woff2;*.WOFF2)",
            "Fichier EOT (*.eot;*.EOT)",
            "Fichier Markdown (*.md;*.markdown;*.MD;*.MARKDOWN)",
            "Fichier GGUF (*.gguf;*.GGUF)"
        ]
        self.comboBox.addItems(file_types)

        self.update_qt_view_from_settings()
        self.radioButton.clicked.connect(self.update_setting_from_qt_view)
        self.radioButton2.clicked.connect(self.update_setting_from_qt_view)
        self.checkBox.clicked.connect(self.update_setting_from_qt_view)
        self.checkBox_new_file.clicked.connect(self.update_setting_from_qt_view)
        self.checkBox_launch_at_startup.clicked.connect(self.update_setting_from_qt_view)


        self.comboBox.currentTextChanged.connect(self.on_text_changed)
        self.pushButton.clicked.connect(self.launch_fucntionalities)
        self.lineEdit.editingFinished.connect(self.update_line_edit_setting)

        if (len(self.line_edit_value)) == 0:
            self.radioButton.setEnabled(True)
            self.radioButton2.setEnabled(True)
            self.comboBox.setEnabled(True)
            self.checkBox_new_file.setEnabled(True)
            if self.checkBox_new_file.isChecked():
                self.checkBox.setEnabled(False)
            else:
                self.checkBox.setEnabled(True)
        else:
            self.radioButton.setEnabled(False)
            self.radioButton2.setEnabled(False)
            self.checkBox.setEnabled(False)
            self.checkBox_new_file.setEnabled(False)
            self.comboBox.setEnabled(False)
        if self.str_Launch_automatically_on_startup=="True":
            self.launch_fucntionalities()

    def update_line_edit_setting(self):
        def strip_quotes(s: str) -> str:
            if len(s) >= 1 and s[0] == s[-1] and s[0] in ("'", '"'):
                return s[1:-1]
            return s
        self.line_edit_value=MetManagement.TransfromPathToStorePath(strip_quotes(str(self.lineEdit.text())))
        if (len(self.line_edit_value)) == 0:
            self.radioButton.setEnabled(True)
            self.radioButton2.setEnabled(True)
            self.comboBox.setEnabled(True)
            self.checkBox_new_file.setEnabled(True)
            if self.checkBox_new_file.isChecked():
                self.checkBox.setEnabled(False)
            else:
                self.checkBox.setEnabled(True)
        else:
            self.radioButton.setEnabled(False)
            self.radioButton2.setEnabled(False)
            self.checkBox.setEnabled(False)
            self.checkBox_new_file.setEnabled(False)
            self.comboBox.setEnabled(False)

    def update_dialog_title_setting(self):
        if getattr(self, 'windowTitleEdit', None) is None:
            return
        value = self.windowTitleEdit.text().strip()
        if self.windowTitleEdit.text() != value:
            self.windowTitleEdit.setText(value)
        self.dialog_window_title = value
        self.update_dialog_title_placeholder()

    def get_default_dialog_title(self):
        if str(self.select_folder_radio_button_checked) == "True":
            return "Select a directory"
        if str(self.new_file) == "True":
            return "Create New File"
        return "Select File(s)"

    def get_effective_dialog_title(self):
        default_title = self.get_default_dialog_title()
        if self.dialog_window_title:
            value = self.dialog_window_title.strip()
            if value:
                return value
        return default_title

    def update_dialog_title_placeholder(self):
        if getattr(self, 'windowTitleEdit', None) is None:
            return
        self.windowTitleEdit.setPlaceholderText(self.get_default_dialog_title())


    @Inputs.input_table
    def input_table(self, data):
        self.in_data = data
        if data is not None:
            self.launch_fucntionalities()



    def launch_fucntionalities(self):
        if self.line_edit_value!="":
            if self.line_edit_value[0]!="?":
                self.selected_path=MetManagement.TransfromStorePathToPath(self.line_edit_value)
            else:
                self.selected_path=self.line_edit_value[1:]
            self.commit_path()
            return
        if str(self.select_folder_radio_button_checked) =="True":
            self.selected_path = self.select_folder(self.get_effective_dialog_title())
            if self.selected_path is None:
                self.selected_path=""
            if self.selected_path=="":
                self.selected_path="error nothing selected"
            self.commit_path()
            return
        multi_select=False
        if str(self.select_multiple_file_checked)!="False":
            multi_select=True
        file_filter=str(self.combo_box_current_text)
        dialog_title = self.get_effective_dialog_title()
        if str(self.new_file)=="False":
            self.selected_path = self.select_file(multi_select, file_filter, dialog_title)
        else:
            self.selected_path = self.select_new_file(file_filter, dialog_title)
        if self.selected_path is None:
            self.selected_path = ""
        if self.selected_path == "":
            self.selected_path = "error nothing selected"
        self.commit_path()
        return

    def select_file(self, multi_select, file_filter, dialog_title):
        if os.name == 'nt':
            return windows_utils.select_file_ctypes(
                multi_select=multi_select,
                file_filter=file_filter,
                dialog_title=dialog_title
            )
        if sys.platform.startswith("Darwin") or sys.platform.startswith("darwin") :
            return  mac_utils.select_file_macos(
                multi_select=multi_select,
                file_filter=file_filter,
                dialog_title=dialog_title
            )
        from AnyQt.QtWidgets import QFileDialog
        files, _ = QFileDialog.getOpenFileNames(
            self,  # parent
            dialog_title,  # window title
            "",  # initial directory
            file_filter  # filter
        )
        files=files[0]
        return files

    def select_new_file(self, file_filter, dialog_title):
        if os.name == 'nt':
            return windows_utils.select_new_file_ctypes(file_filter=file_filter, dialog_title=dialog_title)
        if sys.platform.startswith("Darwin") or sys.platform.startswith("darwin"):
            return mac_utils.select_new_file_macos(file_filter=file_filter, dialog_title=dialog_title)

        from AnyQt.QtWidgets import QFileDialog
        files, _ = QFileDialog.getSaveFileName(
            self,  # parent
            dialog_title,  # window title
            "",  # initial directory
            file_filter  # filter
        )
        return str(files)

    def select_folder(self, dialog_title):
        if os.name=='nt':
            return windows_utils.select_folder_ctypes(title=dialog_title)
        if sys.platform.startswith("Darwin") or sys.platform.startswith("darwin"):
            return mac_utils.select_folder_macos(title=dialog_title)

        # a faire correctement sans qt sur mac
        from AnyQt.QtWidgets import QFileDialog
        return QFileDialog.getExistingDirectory(self, dialog_title)

    def commit_path(self):
        self.error("")
        if self.in_data is not None:
            if "path" in self.in_data.domain or "path" in self.in_data.domain.metas:
                self.error("Path déjà dans les domaines d entrées.")
                self.Outputs.path.send(None)
                self.Outputs.path_string.send(None)

        if not self.selected_path:
            return

        var = StringVariable("path")
        separator = "//////"


        if separator in self.selected_path:
            # plusieurs fichiers

            paths = self.selected_path.split(separator)
            n_paths = len(paths)

            if self.in_data is not None:
                base = self.in_data
                old_dom = base.domain

                # Domaine = attributs + classes existants + nouvelle méta "path"
                domain = Domain(
                    old_dom.attributes,
                    old_dom.class_vars,
                    list(old_dom.metas) + [var]
                )

                m_in = len(base)
                n_rows = max(m_in, n_paths)  # on peut devoir ajouter des lignes
                n_attr = len(domain.attributes)
                n_cls = len(domain.class_vars)
                n_meta = len(domain.metas)

                # Prépare des tableaux de la bonne taille
                X_new = np.full((n_rows, n_attr), np.nan) if n_attr else np.empty((n_rows, 0))
                # Y doit être 2D si des class_vars existent, sinon None
                Y_new = (np.full((n_rows, n_cls), np.nan) if n_cls else None)

                metas_new = np.empty((n_rows, n_meta), dtype=object)
                # Initialiser chaque colonne de méta avec sa valeur "Unknown" Orange
                for j, meta_var in enumerate(domain.metas):
                    unknown = getattr(meta_var, "Unknown", None)
                    metas_new[:, j] = unknown

                # Copie des données existantes dans la partie haute
                if m_in:
                    if n_attr:
                        X_new[:m_in, :] = base.X
                    if n_cls:
                        # Base.Y peut être 1D si une seule classe : on normalise en 2D
                        if base.Y is not None:
                            if np.ndim(base.Y) == 1 and n_cls == 1:
                                Y_new[:m_in, 0] = base.Y
                            else:
                                Y_new[:m_in, :] = base.Y
                    if len(old_dom.metas):
                        metas_new[:m_in, :n_meta - 1] = base.metas  # on laisse la dernière col vide (path)

                # Remplit la nouvelle colonne "path" pour les N premières lignes
                metas_new[:n_paths, -1] = paths

                new_table = Table.from_numpy(domain, X_new, Y_new, metas_new)

            else:
                # Pas de table en entrée : on crée une table uniquement avec la méta "path"
                n_rows = max(n_paths, 1)  # au moins 1 ligne
                domain = Domain([], metas=[var])

                metas_new = np.empty((n_rows, 1), dtype=object)
                metas_new[:, 0] = ""  # vide par défaut
                metas_new[:n_paths, 0] = paths

                X_dummy = np.empty((n_rows, 0))
                new_table = Table.from_numpy(domain, X_dummy, None, metas_new)
            self.Outputs.path.send(new_table)
            self.Outputs.path_string.send(str(paths))
            return
        if self.in_data is not None:
            domain = Domain(
                self.in_data.domain.attributes,
                self.in_data.domain.class_vars,
                list(self.in_data.domain.metas) + [var]
            )
            new_table = Table.from_table(domain, self.in_data)
            new_meta_column = [self.selected_path] * len(new_table)
            new_table.metas[:, -1] = new_meta_column
        else:
            domain = Domain([], metas=[var])
            new_table = Table.from_list(domain, [[]])
            new_table.metas[0] = [self.selected_path]

        self.Outputs.path.send(new_table)
        self.Outputs.path_string.send(self.selected_path)


# Test standalone
if __name__ == "__main__":

    app = QApplication(sys.argv)
    window = OWInputSelector()
    window.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys
import platform
import numpy as np
import datetime as datetime
import Orange
import Orange.data
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import (
    QApplication, QWidget, QCheckBox, QVBoxLayout, QSizePolicy,
    QPushButton, QHBoxLayout, QScrollArea
)

# Conditional imports to mirror Orange add-on layout
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file

try:
    import importlib.metadata as importlib_metadata
except Exception:  # pragma: no cover
    import importlib_metadata  # type: ignore


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWAddColumns(widget.OWWidget):
    name = "Add columns"
    description = "Add selected columns (OS/Python/Orange/packages, etc.) to an input Table."
    category = "AAIT - TOOLBOX"
    icon = "icons/owenvinfo.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owenvinfo.png"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owaddcolumns.ui")
    priority = 1094

    want_main_area = True
    want_control_area = False

    # Settings pour sauvegarder l'état des cases à cocher
    selected_columns = Setting([])  # Liste des noms des colonnes sélectionnées

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    def __init__(self):
        super().__init__()

        # 1) Charger ton .ui d’origine dans un enfant
        self.form: QWidget = uic.loadUi(self.gui)
        self.mainArea.layout().addWidget(self.form)
        self.form.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setMinimumSize(470, 300)

        # 2) Récupérer ce que le .ui pourrait contenir
        self.Description = getattr(self.form, "Description", None)
        self.varsLayout: QVBoxLayout = getattr(self.form, "varsLayout", None)
        self.btnSelectAll = getattr(self.form, "btnSelectAll", None)
        self.btnClearAll = getattr(self.form, "btnClearAll", None)
        self.btnApply = getattr(self.form, "btnApply", None)

        # 3) Si la zone de sélection n'existe pas dans le .ui d’origine, on la crée
        if self.varsLayout is None:
            self._build_fallback_selector_panel()  # crée varsLayout + boutons

        # 4) Etat interne
        self._var_checks: dict[str, QCheckBox] = {}
        self._current_data: Orange.data.Table | None = None

        # 5) Peupler et connecter
        self._available_keys, _ = self._collect()
        self._build_dynamic_checkboxes(self._available_keys)

        if self.btnSelectAll:
            self.btnSelectAll.clicked.connect(self._select_all)
        if self.btnClearAll:
            self.btnClearAll.clicked.connect(self._clear_all)
        if self.btnApply:
            self.btnApply.clicked.connect(self._apply)

    # ---------- Panneau de fallback (si le .ui n'a pas de zone) ----------
    def _build_fallback_selector_panel(self):
        """
        Construit un panneau (boutons + scroll + layout) sous le .ui d’origine,
        et renseigne: self.varsLayout, self.btnSelectAll, self.btnClearAll, self.btnApply.
        """
        panel = QWidget(self)
        panel.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        vroot = QVBoxLayout(panel)
        vroot.setContentsMargins(12, 12, 12, 12)
        vroot.setSpacing(8)

        # Ligne de boutons
        h = QHBoxLayout()
        self.btnSelectAll = QPushButton("Tout cocher", panel)
        self.btnClearAll = QPushButton("Tout décocher", panel)
        h.addWidget(self.btnSelectAll)
        h.addWidget(self.btnClearAll)
        h.addStretch(1)
        self.btnApply = QPushButton("Appliquer", panel)
        h.addWidget(self.btnApply)
        vroot.addLayout(h)

        # Scroll + conteneur de cases à cocher
        scroll = QScrollArea(panel)
        scroll.setWidgetResizable(True)
        scroll.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        inner = QWidget(scroll)
        inner.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.varsLayout = QVBoxLayout(inner)
        self.varsLayout.setContentsMargins(6, 6, 6, 6)
        self.varsLayout.setSpacing(6)
        scroll.setWidget(inner)

        vroot.addWidget(scroll)

        # Ajouter ce panneau APRÈS le .ui d’origine
        self.mainArea.layout().addWidget(panel)

    # ---------- Collecte ----------
    @staticmethod
    def _safe_version(pkg: str):
        try:
            return importlib_metadata.version(pkg)
        except Exception:
            return None

    def _collect(self):
        keys = [
            "OS", "Machine", "Processor",
            "Python Version", "Python Executable", "Orange3",
            "Current Time", "Row number"
        ]
        vals = [
            platform.platform(),
            platform.machine(),
            platform.processor() or platform.machine(),
            sys.version.replace("\n", " "),
            sys.executable,
            getattr(Orange, "__version__", "unknown"),
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        ]

        notable = ["aait", "io4it", "hlit"]
        for pkg in notable:
            ver = self._safe_version(pkg)
            if ver:
                keys.append(pkg)
                vals.append(ver)
        return keys, vals

    # ---------- Cases à cocher dynamiques ----------
    def _build_dynamic_checkboxes(self, keys):
        if not self.varsLayout:
            return  # sécurité (ne devrait plus arriver)

        # Nettoyer
        while self.varsLayout.count():
            item = self.varsLayout.takeAt(0)
            w = item.widget()
            if w:
                w.setParent(None)

        self._var_checks.clear()
        for k in keys:
            cb = QCheckBox(k)
            # Restaurer l'état sauvegardé ou cocher par défaut
            cb.setChecked(k in self.selected_columns if self.selected_columns else True)
            cb.stateChanged.connect(self._update_selected_columns)
            self.varsLayout.addWidget(cb)
            self._var_checks[k] = cb

        self.varsLayout.addStretch(1)

    # ---------- Mise à jour des settings ----------
    def _update_selected_columns(self):
        """Met à jour la liste des colonnes sélectionnées quand une case est cochée/décochée"""
        self.selected_columns = [k for k, cb in self._var_checks.items() if cb.isChecked()]

    # ---------- Slots ----------
    def _select_all(self):
        for cb in self._var_checks.values():
            cb.setChecked(True)

    def _clear_all(self):
        for cb in self._var_checks.values():
            cb.setChecked(False)

    # ---------- Entrée ----------
    @Inputs.data
    def set_data(self, data: Orange.data.Table | None):
        self._current_data = data
        self._apply()

    # ---------- Application ----------
    def _apply(self):
        data = self._current_data
        if data is None:
            self.Outputs.data.send(None)
            return

        keys, vals = self._collect()
        kv = dict(zip(keys, vals))

        # Utiliser les colonnes sélectionnées (sauvegardées dans settings)
        selected_keys = [k for k in self.selected_columns if k in kv] if self.selected_columns else [k for k, cb in
                                                                                                     self._var_checks.items()
                                                                                                     if
                                                                                                     cb.isChecked() and k in kv]

        if not selected_keys:
            self.Outputs.data.send(data)
            return

        existing_meta_vars = list(data.domain.metas)
        new_meta_vars = [Orange.data.StringVariable.make(k) for k in selected_keys]

        new_domain = Orange.data.Domain(
            attributes=list(data.domain.attributes),
            class_vars=list(data.domain.class_vars),
            metas=existing_meta_vars + new_meta_vars,
        )

        n = len(data)
        left = data.metas if (data.metas is not None and data.metas.size) else np.empty((n, len(existing_meta_vars)),
                                                                                        dtype=object)
        right = np.empty((n, len(new_meta_vars)), dtype=object)
        for j, k in enumerate(selected_keys):
            if k == "Row number":
                # Numérotation humaine 1..N (met en 0..N-1 si tu préfères)
                right[:, j] = np.arange(1, n + 1).astype(object)
            else:
                right[:, j] = kv.get(k, "")

        metas = np.hstack([left, right]) if left.size else right

        out = Orange.data.Table(new_domain, data.X, data.Y, metas=metas, W=data.W)
        out.name = data.name or "Data with added columns"
        self.Outputs.data.send(out)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = OWAddColumns()
    w.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()


import os
import sys
import numpy as np

import Orange.data
from AnyQt.QtWidgets import QApplication, QCheckBox, QSpinBox, QComboBox
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Input, Output

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management, base_widget
else:
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from orangecontrib.AAIT.utils import thread_management, base_widget


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWSortAndSelect(base_widget.BaseListWidget):
    name = "Sort & Select"
    description = "This widget sorts the data table on the selected column and keeps the selected number of rows."
    category = "AAIT - TOOLBOX"
    icon = "icons/owsortandselect.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owsortandselect.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owsortandselect.ui")
    want_control_area = False
    priority = 1060

    # Settings
    order = Setting("Ascending")
    limit_rows = Setting(False)
    number_of_rows = Setting(5)

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.data:
            self.var_selector.add_variables(self.data.domain)
            self.var_selector.select_variable_by_name(self.selected_column_name)
        if self.autorun:
            self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(500)
        # Combobox for Ascending or Descending choice
        self.combobox_order = self.findChild(QComboBox, "comboBox")
        self.combobox_order.setCurrentIndex(self.combobox_order.findText(self.order))
        self.combobox_order.currentTextChanged.connect(self.on_order_changed)
        # Checkbox to limit output
        self.checkbox_limit_rows = self.findChild(QCheckBox, "checkBox_2")
        self.checkbox_limit_rows.setChecked(self.limit_rows)
        self.checkbox_limit_rows.toggled.connect(self.toogle_row_limit)
        # Spinbox to choose the number of rows in output
        self.spinbox_number_of_rows = self.findChild(QSpinBox, "spinBox")
        self.spinbox_number_of_rows.setValue(self.number_of_rows)
        self.spinbox_number_of_rows.setEnabled(self.limit_rows)
        self.spinbox_number_of_rows.valueChanged.connect(self.set_row_limit)

        # Data Management
        self.data = None
        self.autorun = True
        self.thread = None
        self.result = None
        self.post_initialized()


    def on_order_changed(self, text):
        self.order = text
        self.run()

    def toogle_row_limit(self, checked):
        self.spinbox_number_of_rows.setEnabled(checked)
        self.limit_rows = checked
        self.run()

    def set_row_limit(self, value):
        self.number_of_rows = value
        self.run()


    def run(self):
        self.warning("")
        self.error("")

        # If Thread is already running, interrupt it
        if self.thread is not None:
            if self.thread.isRunning():
                self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if not self.selected_column_name in self.data.domain:
            self.warning(f'Previously selected column "{self.selected_column_name}" does not exist in your data.')
            self.Outputs.data.send(None)
            return

        if self.checkbox_limit_rows.isChecked():
            number_of_rows = self.number_of_rows
        else:
            number_of_rows = None

        # Start progress bar
        self.progressBarInit()

        # Thread management
        self.thread = thread_management.Thread(sort_table_by_column, self.data, self.selected_column_name, self.order, number_of_rows)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()


    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        self.progressBarFinished()


    def post_initialized(self):
        pass


def sort_table_by_column(table, column_name, order="Ascending", n=None):
    """
    Sort an Orange Table by a column name (feature, class, or meta) and return top n rows.

    Args:
        table (Orange.data.Table): The table to sort.
        column_name (str): Name of the column to sort by.
        order (str): "Ascending" or "Descending".
        n (int or None): Number of rows to return. If None, return all.

    Returns:
        Orange.data.Table: Sorted table with at most n rows.
    """
    var = table.domain[column_name]
    if var is None:
        raise ValueError(f"Column '{column_name}' not found.")

    ascending = False
    if order == "Ascending":
        ascending = True

    # Determine which array to use
    if var in table.domain.attributes:
        values = table.X[:, table.domain.index(var)]
    elif var in table.domain.class_vars:
        values = table.Y
    else:  # meta
        values = table.metas[:, table.domain.metas.index(var)]

    # Sort indices
    if np.issubdtype(values.dtype, np.number):
        sorted_indices = np.argsort(values if ascending else -values, kind="mergesort")
    else:
        sorted_indices = np.argsort(values, kind="mergesort")
        if not ascending:
            sorted_indices = sorted_indices[::-1]

    # Limit number of rows if n is specified
    if n is not None:
        sorted_indices = sorted_indices[:n]

    return table[sorted_indices]

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWSortAndSelect()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()
